"""NaCl SALT2Like error models
"""

import logging

import numpy as np
from scipy import sparse

from bbf.bspline import BSpline, CardinalBSplineC, BSpline2D, lgram
from saltworks import FitParameters
from ...sparseutils import CooMatrixBuff2

try:
    from sksparse.cholmod import cholesky_AAt
except ImportError:
    from scikits.sparse.cholmod import cholesky_AAt


class SimpleErrorSnake:
    """A simplistic error snake : 1 parameter

    ..math ::
      V = g^2 model^2
    """
    def __init__(self, model):
        self.model = model

    def get_struct(self):
        return [("gamma", 1)]

    def init_pars(self, pars):
        pars['gamma'].full[:] = 0.01
        
    def __call__(self, pars, jac=False):
        v, J = None, None
        if jac:
            v, J = self.model(pars, jac=True)
        else:
            v = self.model(pars, jac=False)
        N = len(v)
        g = pars['gamma'].full[0]
        var = np.full(N, g**2 * v**2)

        if not jac:
            return var

        J = J.tocoo()

        n_free_pars = len(pars.free)
        i = [J.row]
        j = [J.col]
        data = [2. * g**2 * v[J.row] * J.data]
        i.append(np.arange(N))
        j.append(np.full(N, pars['gamma'].indexof(np.zeros(N).astype(int))))
        data.append(2. * g * v**2)

        i = np.hstack(i)
        j = np.hstack(j)
        data = np.hstack(data)

        idx = j >= 0
        J = sparse.coo_matrix((data[idx], (i[idx], j[idx])),
                              shape=(N, n_free_pars))

        return var, J


class LocalErrorSnake:
    """Error snake variance - as a function of phase and wavelength

    .. math::
    V(\lambda, p) = \sigma_{Err}^2 + \sigma_{\mathrm{Mod}}^2(\lambda,p)

    where :

    .. math::
        \sigma_{\mathrm{Mod}}^{spec} &= g(\lambda, p) \times \phi_{\mathrm{spec}}(p,\lambda)\ \ (\mathrm{spectra}) \\
        \sigma_{\mathrm{Mod}}^{phot} &= g(\lambda, p) \times \phi_{\mathrm{phot}}(p,\lambda)\ \ (\mathrm{photometric})

    :math:`\gamma(\lambda, p)` is a global surface describing the spectral
    residual of a 2D spline surface defined on the same ranges as the
    corresponding model.
    """
    def __init__(self, model, bins=(10, 10), order=4):
        """Constructor
        """
        self.model = model
        self.training_dataset = model.training_dataset

        # instantiate a basis
        ph_bins, wl_bins = bins
        phase_grid = np.linspace(model.basis.by.range[0], model.basis.by.range[-1], ph_bins)
        wl_grid = np.linspace(model.basis.bx.range[0], model.basis.bx.range[-1], wl_bins)
        self.basis = BSpline2D(wl_grid, phase_grid, x_order=order, y_order=order)

    def get_struct(self):
        """
        """
        return [('gamma', len(self.basis))]

    def init_pars(self, pars):
        """
        """
        pars['gamma'].full[:] = 0.01

    def __call__(self, pars, jac=False):
        """evaluate the variance model, and optionnaly, its derivatives
        """
        # too bad we need to call the model again here...
        # let's see whether we could cache this call
        if jac:
            flux, Jm = self.model(pars, jac=True)
        else:
            flux = self.model(pars, jac=False)
            Jm = None

        tds = self.training_dataset
        zz = tds.lc_data.z
        tmax = pars['tmax'].full[tds.lc_data.sn_index]
        wl = tds.lc_data.wavelength / (1. + zz)
        phase = (tds.lc_data.mjd - tmax) / (1. + zz)
        Jv = self.basis.eval(wl, phase)

        g = np.zeros(len(tds))
        g[:len(tds.lc_data)] = np.exp(Jv @ pars['gamma'].full)
        v = g * flux**2
        if not jac:
            return v

        # derivatives
        N = len(tds)
        n_free_pars = len(pars.free)
        buff = CooMatrixBuff2((N, n_free_pars))

        # d_gamma
        buff.append(Jv.row,
                    pars['gamma'].indexof(Jv.col),
                    g[Jv.row] * Jv.data * flux[Jv.row]**2)

        # d_dpars
        buff.append(Jm.row,
                    Jm.col,
                    2. * g[Jm.row] * flux[Jm.row] * Jm.data)

        J = buff.tocoo()

        return v, J



class SNLambdaErrorSnake:

    def __init__(self, model, deg=3):
        self.model = model
        self.training_dataset = self.model.training_dataset
        self.deg = deg
        
        #pre construction of the vandermonde matrix
        lc_data = self.training_dataset.lc_data
        restframe_wl = lc_data.wavelength / (1. + lc_data.z)
        r_wl = np.zeros(len(self.training_dataset))
        r_wl[:len(lc_data)] = restframe_wl
        self.vander_matrix = self.get_vander_matrix(r_wl)

    def get_struct(self):
        nsn = len(self.training_dataset.sn_data.sn_set)
        npars = self.deg # deg + 1 - 1 constraint
        return [('gamma_sn', nsn), ('gamma_snake', npars)]

    def init_pars(self, pars):
        pars['gamma_sn'].full[:] = 0.01
        pars['gamma_snake'].full[:] = 0.

    def _reduce(self, restframe_wl):
        return (restframe_wl-4500.) / (9000. - 2000.)
    
    def get_vander_matrix(self, restframe_wl):
        reduced_wl = self._reduce(restframe_wl)
        vander_mat = np.vander(reduced_wl, self.deg, increasing = False)
        return vander_mat
    
    def __call__(self, pars, jac=False):
        if jac:
            flux, Jm = self.model(pars, jac=True)
        else:
            flux = self.model(pars, jac=False)
            Jm = None

        tds = self.training_dataset
        zz = tds.lc_data.z
        restframe_wl = tds.lc_data.wavelength / (1. + zz)
        reduced_wl = self._reduce(restframe_wl)
        snake = np.zeros(len(tds))
        snake[:len(tds.lc_data)] = np.exp(np.polyval(pars['gamma_snake'].full, reduced_wl))

        g = np.zeros(len(tds))
        g[:len(tds.lc_data)] = pars['gamma_sn'].full[tds.lc_data.sn_index]
        variance = g**2 * snake * flux**2
        
        if not jac:
            return variance
        
        # derivatives
        N = len(tds)
        n_free_pars = len(pars.free)
        buff = CooMatrixBuff2((N, n_free_pars))
        
        # d_gamma = 2 * snake * g_i * flux**2
        buff.append(tds.lc_data.row,
                    pars['gamma_sn'].indexof(tds.lc_data.sn_index),
                    2 * snake[tds.lc_data.row] * g[tds.lc_data.row] * flux[tds.lc_data.row]**2)
        
        # d_snake = g**2 * flux**2 * reduced_wl ** i * snake
        data_d_snake = sparse.coo_matrix(variance[:,np.newaxis] * self.vander_matrix[:,0:self.deg-1])
        row_d_snake = data_d_snake.row
        col_d_snake = pars['gamma_snake'].indexof(data_d_snake.col)
        buff.append(row_d_snake,
                    col_d_snake,
                    data_d_snake.data)
        
        # d_pars = 2 * snake * g**2 + flux * Jm
        buff.append(Jm.row,
                    Jm.col,
                    2. * g[Jm.row]**2 * snake[Jm.row] * flux[Jm.row] * Jm.data)

        J = buff.tocoo()

        return variance, J


class SNLocalErrorSnake:
    """Error snake variance - as a function of phase and wavelength

    .. math::
    V(\lambda, p) = \sigma_{Err}^2 + \sigma_{\mathrm{Mod}}^2(\lambda,p) \times \sigma_{\mahtrm{SN}}

    where :

    .. math::
        \sigma_{\mathrm{Mod}}^{spec} &= g(\lambda, p) \times \gamma_{\mahtrm{SN}} \times \phi_{\mathrm{spec}}(p,\lambda)\ \ (\mathrm{spectra}) \\
        \sigma_{\mathrm{Mod}}^{phot} &= g(\lambda, p) \times \gamma_{\mahtrm{SN}} \times \phi_{\mathrm{phot}}(p,\lambda)\ \ (\mathrm{photometric})

    :math:`\gamma(\lambda, p)` is a global surface describing the spectral
    residual of a 2D spline surface defined on the same ranges as the
    corresponding model.
    :math:`\gamma_{\mahtrm{SN}}` is a de-weighting factor for each SN to capture
    SN variability
    """
    def __init__(self, model, bins=(10, 10), order=4):
        """Constructor
        """
        self.model = model
        self.training_dataset = model.training_dataset

        # instantiate a basis
        ph_bins, wl_bins = bins
        phase_grid = np.linspace(model.basis.by.range[0], model.basis.by.range[-1], ph_bins)
        wl_grid = np.linspace(model.basis.bx.range[0], model.basis.bx.range[-1], wl_bins)
        self.basis = BSpline2D(wl_grid, phase_grid, x_order=order, y_order=order)

    def get_struct(self):
        """
        """
        nsn = len(self.training_dataset.sn_data.sn_set)
        return [('gamma_snake', len(self.basis)), ('gamma_sn', nsn)]

    def init_pars(self, pars):
        """
        """
        pars['gamma_sn'].full[:] = 0.01
        pars['gamma_snake'].full[:] = 0.01

    def __call__(self, pars, jac=False):
        """evaluate the variance model, and optionnaly, its derivatives
        """
        if jac:
            flux, Jm = self.model(pars, jac=True)
        else:
            flux = self.model(pars, jac=False)
            Jm = None

        tds = self.training_dataset
        zz = tds.lc_data.z
        tmax = pars['tmax'].full[tds.lc_data.sn_index]
        wl = tds.lc_data.wavelength / (1. + zz)
        phase = (tds.lc_data.mjd - tmax) / (1. + zz)
        Jv = self.basis.eval(wl, phase)
        
        # fixing phase edges parameters
        njx = self.basis.bx.nj
        nj = self.basis.nj
        cols_to_keep = np.arange(0, nj-njx-1, 1)
        
        #cols_full = np.arange(0, nj, 1)
        #b = np.array([0, njy -1])
        #c = np.arange(0, njy, 1)
        #a = c*njy
        #x = [b+i for i in a]
        #x = np.hstack(x)
        #cols_remove = np.isin(cols_full, x)
        #cols_to_keep = ~cols_remove
        
        mask = np.isin(Jv.col, cols_to_keep)
        
        filtered_data = Jv.data[mask]
        filtered_row = Jv.row[mask]
        filtered_col = Jv.col[mask]
        
        col_map = {old_col:new_col for new_col, old_col in enumerate(cols_to_keep)}
        filtered_col = np.array([ col_map[c] for c in filtered_col ])
        new_shape = (Jv.shape[0], len(cols_to_keep))
        Jv = sparse.coo_matrix( (filtered_data, (filtered_row, filtered_col)), shape=new_shape)
        
        #
        g_snake = np.zeros(len(tds))
        g_snake[:len(tds.lc_data)] = np.exp(Jv @ pars['gamma_snake'].full[:-njx-1])
        
        g_sn = np.zeros(len(tds))
        g_sn[:len(tds.lc_data)] = pars['gamma_sn'].full[tds.lc_data.sn_index]
        
        v = g_snake * g_sn**2 * flux**2
        if not jac:
            return v

        # derivatives
        N = len(tds)
        n_free_pars = len(pars.free)
        buff = CooMatrixBuff2((N, n_free_pars))

        # d_gamma_snake
        buff.append(Jv.row,
                    pars['gamma_snake'].indexof(Jv.col),
                    g_snake[Jv.row] * Jv.data * g_sn[Jv.row]**2 * flux[Jv.row]**2)
        
        # d_gamma_sn
        buff.append(tds.lc_data.row,
                    pars['gamma_sn'].indexof(tds.lc_data.sn_index),
                    2 * g_sn[tds.lc_data.row] * g_snake[tds.lc_data.row] * flux[tds.lc_data.row]**2)
        
        # d_dpars
        buff.append(Jm.row,
                    Jm.col,
                    2. * g_snake[Jm.row] * g_sn[Jm.row]**2 * flux[Jm.row] * Jm.data)

        J = buff.tocoo()

        return v, J


class SimpleErrorSnake_:
    """Simple, 1 parameter error snake: :math:`\\sigma_f = \\sigma_m \\times f`

    This is the simplest error snake possible. We add a pedestal variance,
    constant in magnitude (i.e. of the form :math:`\\sigma_m \times flux` in
    flux).
    """

    def __init__(self, model):  # , var_pedestal=0.001):
        self.model = model
        self.training_dataset = self.model.training_dataset
        # self.var_pedestal = var_pedestal

    def get_struct(self):
        """
        """
        return [('sigma_snake', 1)]

    def __call__(self, pars, jac=False, model_flux=None, model_jac=None):
        """return the diagonal of the error snake variance + derivatives
        """
        mflux, mjac = None, None
        # if parameters are given, we need to re-evaluate the model
        if p is not None:
            if not jac:
                mflux = self.model(p)
            else:
                mflux, mjac = self.model(p, jac=True)
        # otherwise, we re-evaluate it only if it is necessary
        else:
            if not jac:
                mflux = model_flux if model_flux is not None else self.model(self.model.pars.free)
            else:
                if model_flux is not None and model_jac is not None:
                    mflux, mjac = model_flux, model_jac
                else:
                    mflux, mjac = self.model(self.model.pars.free, jac=True)

        # model parameters
        pars = self.model.pars
        sigma_m = self.model.pars['sigma_snake'].full[0]

        # evaluate the variance
        var = (sigma_m * mflux)**2

        # we should only apply it to the photometric measurements
        # nlc, _, _ = self.model.training_dataset.n

        # and return it if this is the only thing requested
        if not jac:
            return var

        # derivatives
        n = len(self.model.pars.full)
        n_free = len(self.model.pars.free)
        N = self.training_dataset.nb_meas(valid_only=False)

        # dV / dsigma_snake
        i = np.arange(N).astype(int)
        j = np.full(N, pars['sigma_snake'].indexof())
        dVds = 2. * sigma_m * mflux**2
        idx = j>=0
        J_dVds = sparse.coo_matrix((dVds[idx], (i[idx],j[idx])), shape=(N,n_free))

        # dV/dbeta
        # mjac.data *= (2. * sigma_m**2 * mflux[mjac.row])
        J_dVdbeta = sparse.coo_matrix((mjac.data * 2. * sigma_m**2 * mflux[mjac.row],
                                (mjac.row, mjac.col)),
                                shape=mjac.shape)

        # i = np.hstack((i, mjac.row))
        # j = np.hstack((j, mjac.col))
        # v = np.hstack((dVds, mjac.data))
        # idx = j>=0
        # J = coo_matrix((v[idx], (i[idx],j[idx])), shape=(N, n_free))
        # J = J_dVds.tocsc() + J_dVdbeta.tocsc()
        J = J_dVds.tocoo() + J_dVdbeta.tocoo()

        return var, J

    def noise(self, p=None, **kwargs):
        """draw one realization of the error snake noise
        """
        var = self(p, jac=False, **kwargs)
        nlc, nsp, nphotspec = self.training_dataset.nb_meas(valid_only=False,
                                                            split_by_type=True)
        N = nlc + nsp + nphotspec
        n = np.zeros(N)
        # this error snake only applies to the photometric dataset
        n[:nlc] = np.random.normal(loc=0., scale=np.sqrt(var[:nlc]))
        return n


class CalibErrorModel:

    def __init__(self, model, calib_variance, 
                 default_calib_variance=0.5**2):
        """Constructor

        Parameters
        ----------
        model : a NaCl model (e.g. SALT2Like)
            the underlying model
        calib_variance : float or pandas.DataFrame
            the calibration covariance matrix. It is either a float, if all
            bands have the same calibration variance (uncorrelated), or if all
            calibration uncertainties are uncorrelated, the diagonal of the
            covariance matrix, or the full covariance matrix.
        """
        self.model = model
        self.training_dataset = self.model.training_dataset
        self.calib_variance = calib_variance
        self.default_calib_variance = default_calib_variance
        self.inv_full_covmat = None
        self.sub_covmat = None

    # def finalize(self):
    #     """
    #     """
    #     self.build_covmat(calib_variance=self.calib_variance,
    #                       default_calib_variance=self.default_calib_variance)

    def get_struct(self):
        nb_bands = self.training_dataset.nb_bands()
        return [('eta_calib', nb_bands)]

    def init_pars(self, pars):
        pass

    def build_covmat(self, pars,
                     calib_variance, bands=None,
                     default_calib_variance=0.5):
        """build a covmat (ordered to match the lc_data.band_index)

        Returns
        -------
        np.ndarray
            the re-ordered covmat
        """
        npars = len(pars.full)
        pfull = pars.copy()
        pfull.release()

        sz = len(pfull['eta_calib'].full)

        # if all bands have the same variance, that's easy
        if isinstance(calib_variance, float):
            nb_bands = self.training_dataset.nb_bands()
            i = pfull['eta_calib'].indexof()
            v = np.full(sz, 1./self.calib_variance)
            self.sub_covmat = np.diag(np.full(nb_bands, calib_variance))
            self.inv_full_covmat = sparse.coo_matrix((v, (i, i)), shape=(npars, npars))

        # otherwise, we may get a dict of the form {<band_name>: var, ...}
        # in which case it is also easy
        elif isinstance(calib_variance, dict):
            tds = self.model.training_dataset
            diag = []
            for bn in tds.lc_data.band_set:
                if bn not in calib_variance:
                    logging.warning(f'no calibration variance specified for band: {bn} -- will assign {default_calib_variance} by default')
                diag.append(calib_variance.get(bn, default_calib_variance))
            diag = np.array(diag)
            assert np.all(diag)
            i = pfull['eta_calib'].indexof()
            self.sub_covmat = np.diag(diag)
            self.inv_full_covmat = sparse.coo_matrix((1./diag, (i,i)), shape=(npars, npars))

        # finally, we may get a real matrix, along with a dict that
        # specifies the indices to band mapping. Then, all we have
        # to do, is to make sure all the bands are here, fill out the blank
        # if necessary, and re-order it if necessary
        elif isinstance(calib_variance, tuple):
            covmat, bands  = calib_variance
            # extend band into a simple map
            band_map = dict([(y,x) for x,y in enumerate(bands)])

            # extend calib covmat with a defaut value
            nr,nc = covmat.shape
            assert nr == nc
            extended_covmat = np.zeros((nr+1,nc+1))
            extended_covmat[:nr,:nc] = covmat
            extended_covmat[nr,nr] = default_calib_variance
            #            print('extended_covmat: ', extended_covmat)
            # re-order covmat
            i = []
            for bn in self.model.training_dataset.lc_data.band_set:
                if bn not in bands:
                    logging.warning('no calibration variance specified for band: {bn} -- will assign {default_calib_variance} by default')
                    i.append(nr)
                else:
                    i.append(band_map[bn])
            ii,jj = np.meshgrid(i,i)
            reordered_covmat = extended_covmat[ii,jj]

            # invert the covmat 
            w = np.linalg.inv(reordered_covmat)

            # and re-index it into a larger (n,n) matrix, where n = number of parameters
            nn, _ = w.shape
            ii, jj = np.mgrid[:nn,:nn]
            self.sub_covmat = reordered_covmat
            self.inv_full_covmat = sparse.coo_matrix((w.flatten(),
                                               (pfull['eta_calib'].indexof(ii.flatten()), 
                                                pfull['eta_calib'].indexof(jj.flatten()))), 
                                              shape=(npars, npars))
        else:
            raise ValueError(f"build_covmat(): don't know what to do with {calib_variance}")

        # # otherwise, if a matrix is passed, we may need to re-order
        # # the calibration matrix, so that it matches the band_index
        # if isinstance(calib_variance, np.matrix):
        #     C = coo_matrix(calib_variance)
        #     assert (len(C.shape) == 2) and (C.shape[0] == C.shape[1])

        # if isinstance(self.calib_variance, float):
        #     n_bands = self.training_dataset.nb_bands()
        #     return np.diag(np.full(n_bands, self.calib_variance))

        # if isinstance(self.calib_variance, pandas.DataFrame) and self.calib_variance.shape[0] == 1:
        #     diag = self.calib_variance[self.lc_data.band_set].flatten()
        #     return np.diag(diag)

        # if isinstance(self.calib_variance, pandas.DataFrame):
        #     bs = self.lc_data.band_set
        #     covmat = self.calib_variance[bs].loc[bs].to_numpy()
        #     return covmat

    def __call__(self, pars, deriv=False):
        """evaluate the calib prior (and optionally its gradient and Hessian)
        """
        #        if p is not None:
        #            self.model.pars.free = p

        if self.sub_covmat is None or self.inv_full_covmat is None:
            self.build_covmat(pars, self.calib_variance)

        pfull = pars.full
        prior = np.dot(pfull, (self.inv_full_covmat.dot(pfull)))
        if not deriv:
            return prior

        # TODO: add some sanity and change the sign in the LogLikelihood
        # so that can return grad and not -grad ...
        # TODO: I am here ! -> reduce size, depending on fixed parameters
        grad = -2. * self.inv_full_covmat.dot(pfull)
        hess = +2. * self.inv_full_covmat

        return prior, grad, hess

    def correction(self, pars, lcdata=None):
        """the multiplicative correction to apply to the photometric data

        Parameters
        ----------
        p: (ndarray), optional
          vector of free parameters
        lcdata: stff, optional
          something that has a band_index

        Returns:
          a ndarray with the corrections
        """
#        if p is not None:
#            self.model.pars.free = p
        if lcdata is None:
            lcdata = self.training_dataset.lc_data
        corr = 1. + pars['eta_calib'].full[lcdata.band_index]
        return corr

    def noise(self, pars):
        """the noise component, to apply to the photometric data
        """
        # if p is not None:
        #     self.model.pars.free = p

        lc_data = self.training_dataset.lc_data

        nphot, nsp, nspphot = self.training_dataset.nb_meas(valid_only=False,
                                                            split_by_type=True)
        scales = np.ones(nphot + nsp + nspphot)
        nb_bands = self.training_dataset.nb_bands()

        # covmat = self.build_covmat()
        L = np.linalg.cholesky(self.sub_covmat)
        dm = L @ np.random.normal(0., scale=1., size=nb_bands)
        # dm = [dm[self.band_map[i]] for i in range(len(self.bands))]
        scales[:nphot] = 1. + dm[lc_data.band_index]
        return scales # 1.+ dm[lc_data.band_index]


class ColorScatterModel:
    """An adaptation of the color scatter model developed by Guy Augarde
    """
    U_WAVELENGTH = 3650.88
    B_WAVELENGTH = 4302.57
    V_WAVELENGTH = 5428.55
    R_WAVELENGTH = 6418.01
    I_WAVELENGTH = 7968.34
    WAVELENGTH = {"U": U_WAVELENGTH, "B": B_WAVELENGTH, "V": V_WAVELENGTH,
                  "R": R_WAVELENGTH, "I": I_WAVELENGTH}

    def __init__(self, model):
        """Constructor
        """
        self.model = model
        self.training_dataset = model.training_dataset
        lc_data = self.training_dataset.lc_data
        restframe_wl = lc_data.wavelength / (1. + lc_data.z)
        self.reduced_restframe_wl = self.model.reduce(restframe_wl)

    def get_struct(self):
        """
        """
        nb_lightcurves = self.training_dataset.nb_lcs()
        return [('sigma_kappa', 3),
                ('kappa_color', nb_lightcurves)]

    def __call__(self, p=None, deriv=False):
        """return the color scatter prior, and (optionaly) its grad and hessian
        """
        pass

    def correction(self, pars):
        """return the multiplicative correction to be applied to the model
        """
        # if p is not None:
        #    self.model.pars.free = p
        tds = self.training_dataset
        corr = 1. + pars['kappa_color'].full[tds.lc_data.lc_index]
        return corr

    def noise(self, pars):
        """generate one realization of the noise -- to add directly to the data
        """
        # if p is not None:
        #     self.model.pars.free = p
        var = self.__call__(pars, jac=False)
        nn = np.random.normal(scale=np.sqrt(var), size=len(var))
        return 1. * nn[self.training_dataset.lc_data.lc_index]


class ColorScatter:
    r"""color scatter model

    To model the residual variability of the SNe color not described by the
    model, we allow the relative amplitude of each light curve to vary by a
    quantity :math:`(1+kappa)`:

    .. math::
        \phi_{phot}(p_0) = X_0 (1+z) (1+kappa) \int S(\lambda, p) T\left( \lambda \right) \frac{\lambda}{hc} d\lambda

    where :math:`\kappa` is an additional parameter, depending on the SN and the
    observation band.

    The variability of :math:`\kappa` is defined by a Gaussian prior, whose
    variance is a function of the wavelength restframe, and must be determined
    during the adjustment. In our model, this term is implemented by a
    polynomial of the wavelength restframe of the SN. For a light curve observed
    in a band :math:`X`, of average wavelength :math:`\lambda^X`, the
    corresponding variance is:

    .. math::
        \sigma_\kappa^2 = P\left(\frac{\lambda^X}{1+z}\right)

    where :math:`z` is redshift of the supernova and :math:`P` is a polynomial
    of the wavelength restframe. In practice, :math:`P` is implemented in terms
    of a reduced restframe wavelength: :math:`\lambda_{r}`

    Attributes
    ----------
    WL_REDUCED : numpy.array
        Reduced SN-restframe mean wavelength
    pars : nacl.lib.fitparameters.FitParameters
        Color scatter parameters
    """
    U_WAVELENGTH = 3650.88
    B_WAVELENGTH = 4302.57
    V_WAVELENGTH = 5428.55
    R_WAVELENGTH = 6418.01
    I_WAVELENGTH = 7968.34
    WAVELENGTH = {"U": U_WAVELENGTH, "B": B_WAVELENGTH, "V": V_WAVELENGTH,
                  "R": R_WAVELENGTH, "I": I_WAVELENGTH}

    def __init__(self, model): # wavelength_rest_reduced, sigma_kappa):
        """Constructor

        Parameters
        ----------
        wavelength_rest_reduced :
            Reduced SN-restframe mean wavelength
        sigma_kappa : numpy.array
            Color scatter parameter initialisation.
        """
        self.model = model
        self.training_dataset = model.training_dataset
        lc_data = self.training_dataset.lc_data
        restframe_wl = lc_data.wavelength / (1. + lc_data.z)
        self.reduced_restframe_wl = self.model.reduce(restframe_wl)

        # self.WL_REDUCED = wavelength_rest_reduced
        # self.pars = self.init_pars(sigma_kappa)

    def get_struct(self):
        """
        """
        nb_lightcurves = self.training_dataset.nb_lcs()
        return [('sigma_kappa', 3),
                ('kappa_color', nb_lightcurves)]

    # def init_pars(self):
    #     fp = FitParameters(self.get_struct())

    # @staticmethod
    # def init_pars(sigma_kappa):
    #     """initialize a parameter vector

    #     Parameters
    #     ----------
    #     sigma_kappa : numpy.array
    #         Color scatter parameter initialisation.

    #     Returns
    #     -------
    #     pars : nacl.lib.fitparameters.FitParameters
    #         Color scatter parameter.
    #     """
    #     pars = FitParameters([('sigma_kappa', len(sigma_kappa))])
    #     pars['sigma_kappa'] = sigma_kappa
    #     return pars

    def __call__(self, pars, jac=False):
        """Return the color scatter variance (the :math:`V_\kappa` matrix)

        Parameters
        ----------
        p : np.ndarray
          free parameter vector

        Returns
        -------
        val : numpy.ndarray
            the diagonal of the color scatter variance matrix (used in the kappa prior)
        if jac:
            jacobian of the color scatter.
        """
        # if p is not None:
        #     self.model.pars.free = p
        sigma_kappa = pars['sigma_kappa'].full
        val = np.polyval(sigma_kappa, self.reduced_restframe_wl)  # * self.WL_REDUCED

        if jac:
            vander = (np.vander(self.reduced_restframe_wl, len(sigma_kappa)).T )  # * self.WL_REDUCED)
            jj = (vander * val * 2).T
            return val, jj  # coo_matrix(jj)
        return val

    def correction(self, pars):
        """return the multiplicative correction to be applied to the model
        """
        # if p is not None:
        #     self.model.pars.free = p
        corr = 1. + pars['kappa_color'].full[self.training_dataset.lc_data.lc_index]
        return corr

    def noise(self, pars):
        """generate one realization of the noise -- to add directly to the data
        """
        # if p is not None:
        #     self.model.pars.free = p
        var = self.__call__(pars, jac=False)
        nn = np.random.normal(scale=np.sqrt(var), size=len(var))
        return nn[self.training_dataset.lc_data.lc_index]



class SimpleVarianceModel:
    """Error snake variance

    The error snake variance describes the excess of variance observed in the
    training residuals, that arise from the fact that the model is not able to
    entirely capture the SN-to-SN variability.

    In NaCl, this component incorporated in the fit, and determined along with
    the model itself.

    .. math::
    V(\lambda, p) = \sigma_{Err}^2 + \sigma_{\mathrm{Mod}}^2(\lambda,p)

    where :

    .. math::
        \sigma_{\mathrm{Mod}}^{spec} &= g(\lambda, p) \times \phi_{\mathrm{spec}}(p,\lambda)\ \ (\mathrm{spectra}) \\
        \sigma_{\mathrm{Mod}}^{phot} &= g(\lambda, p) \times \phi_{\mathrm{phot}}(p,\lambda)\ \ (\mathrm{photometric})

    :math:`\gamma(\lambda, p)` is a global surface describing the spectral
    residual of a 2D spline surface defined on the same ranges as the
    corresponding model.


    Attributes
    ----------
    disconnect_sp : bool
        Whether error snake of spectral data is considered.
    model : nacl.salt
        Model.
    sp : nacl.dataset.SpectrumData
        Spectral data
    lc : nacl.dataset.LcData
        Photometric data
    filter_db : nacl.instruments.FilterDb
        Filter transmission projected on B-splines
    bands : numpy.array
        Bands used for photometric data
    threshold : int
        Fix the spline parameters for bin that gather less data points than this
        threshold value.
    wl_grid : numpy.array
        Wavelength grid.
    phase_grid : numpy.array
        Phase grid.
    basis : nacl.lib.bspline.BSpline2D
        2D spline basis.
    factor_int : float
        Model normalisation.
    lambda_c : numpy.array
        Mean wavelength of the filter in the SN-restframe.
    pars : nacl.lib.fitparameters.FitParameters
        Parameters :math:`\gamma`.
    pars0 : numpy.array
        Parameters used to initialize the error snake.
    """
    def __init__(self, model, gamma_init=0.01,
                #  disconnect_sp=False,
                threshold=None,
                #  adaptive_grid=False,
                 bins=(3, 3), order=4):
        """Constructor

        - instantiate the bases (phase, wavelength)
        - compute the \lambda_c for each light curve
        - initiate the parameters

        Parameters
        ----------
        model : nacl.salt
            Model.
        gamma_init : numpy.array ou float
            Parameters used to initialize the error snake.
        disconnect_sp : bool
            Whether error snake of spectral data is considered.
        threshold : int
            Fix the spline parameters for bin that gather less data points than this threshold value.
        adaptive_grid : bool
            It true, define Bspline knots where parameter space has data.
        bins : tuple
            Number of bins if the grid is not adaptive.
        order :
            Spline order.
        """
        # no reason to apply this to the spectral data
        # self.disconnect_sp = disconnect_sp

        self.model = model
        self.training_dataset = model.training_dataset
        self.filter_db = self.model.filter_db
        self.bands = self.model.bands
        self.threshold = threshold
        order = order

        lc_data = self.training_dataset.lc_data
        sp_data = self.training_dataset.spec_data

        # if we need an adaptive grid, we'll reconnect that later
        # if adaptive_grid:
        #     phase_lc = self.model.get_restframe_phases(self.lc)
        #     phase_sp = self.model.get_restframe_phases(self.sp)
        #     wavelength_lc = lc_data.wavelength/(1. + lc_data.z)
        #     wavelength_sp = sp_data.wavelength/(1. + sp_data.z)
        #     phase = np.hstack((phase_sp, phase_lc))
        #     wavelength = np.hstack((wavelength_sp, wavelength_lc))
        #     _, wl_grid, phase_grid, _ = pl.hist2d(wavelength, phase, bins=2)
        # else:
            # print(f'\n VarModel N bins : ph : {ph_bin}, wl {wl_bin}')
        ph_bins, wl_bins = bins
        phase_grid = np.linspace(model.basis.by.range[0], model.basis.by.range[-1], ph_bins)
        wl_grid = np.linspace(model.basis.bx.range[0], model.basis.bx.range[-1], wl_bins)
        self.basis = BSpline2D(wl_grid, phase_grid, x_order=order, y_order=order)

        # not sure what this is...
        # Guy tells me it is not used anymore.
        # self.factor_int = self.model.norm

        # this should not happen. If it does, this is certainly not the
        # responsibility of a subcomponent of the error model to correct for
        # this.
        # if lc_data.wavelength.sum() == 0:
        #     filter_mean_wl = [self.model.filter_db.transmission_db[bd].mean() for bd in self.model.bands.astype(str)]
        #     self.lambda_c = np.array(filter_mean_wl)[self.lc['band_id']]/(1+self.lc['ZHelio'])
        # else:
        self.lambda_c = lc_data.wavelength / (1.+lc_data.z) # ['Wavelength']/(1+self.lc['ZHelio'])
        self.pars = self.init_pars(gamma_init=gamma_init)
        # self.pars0 = self.pars.full.copy() # ?

    def fix_non_constrained_pars(self, threshold):
        """index of parameters of the parameter space that don't have enough data to constraint the error snake.

        Parameters
        ----------
        threshold : int
            Threshold value defining the "enough". If the sum of the data in a bin is less than this value,
            the corresponding parameter is fixed.

        Returns
        ----------
        idx : numpy.array
            Index of parameters that will be fixed.
        """
        lc_data = self.training_dataset.lc_data
        # sp_data = self.training_dataset.spec_data
        lc_phases = self.model.get_restframe_phases(lc_data)
        # sp_phases = self.model.get_restframe_phases(sp_data)
        lc_wl = lc_data.wavelength / (1. + lc_data.z)
        # sp_wl = sp_data.wavelength / (1. + sp_data.z)

        # x = np.hstack((wl_sp, wl_lc))
        # y = np.hstack((ph_sp, ph_lc))
        jj = self.basis.eval(lc_wl, lc_phases)
        idx = np.where(np.bincount(np.sort(jj.col)) < threshold)
        return idx[0]

    def init_pars(self, gamma_init):
        """
        Initiate model parameters.

        Parameters
        ----------
        gamma_init : float
            Parameters used to initialize the error snake.

        Returns
        ----------
        gamma : nacl.lib.fitparameters.FitParameters
            Parameters :math:`\gamma`.
        """
        gamma = FitParameters([('gamma', self.basis.bx.nj * self.basis.by.nj)])
        gamma['gamma'] = gamma_init
        if self.threshold is not None:
          idx = self.fix_non_constrained_pars(self.threshold)
          gamma.fix(idx)
        return gamma
        #pass

    def get_struct(self):
        npars = self.basis.bx.nj * self.basis.by.nj
        return [('gamma', npars)]

    def noise(self, p=None):
        lc_data = self.training_dataset.lc_data
        var = self(p=None)
        noise = np.random.normal(scale=var, size=len(lc_data))
        return noise

    def __call__(self, pars, jacobian=None, debug_mode=False):
        """evaluate the model - return a variance and (optionally) its derivatives

        Evaluate the model for a set of parameter :math:`\gamma` (variance model parameters)
        and :math:`\beta` (flux model parameters).

        Parameters
        ----------
        gamma : np.array
           set of parameter of the variance model;
        beta : np.array
           set of parameter of the flux model;
        jacobian : None or scipy.sparse.coo_matrix
           derivative of the flux, not necessary if derivative of variance model not needed,
        debug_mode : bool
           to get some variables...
        lc : nacl.dataset.LcData
           to get variance model of a set of light curve and is None will get for all
        sp : nacl.dataset.SpectrumData
           to get variance model of a set of spectra and is None will get for all

        Returns
        ----------
        var : numpy.array
            Variance value of the model, quadratic sum of the error measurement and the error snake.
        if jacobian is not None:
        var : numpy.array
            Variance value of the model, quadratic sum of the error measurement and the error snake.
        jacobian_var : scipy.sparse.coo_matrix
            Jacobian (derivatives wrt :math:`\beta` and :math:`\gamma`)matrix of the error model.
        """
        # self.model.pars.free = p

        lc_data = self.training_dataset.lc_data
        lc_phases = self.model.get_restframe_phases(lc_data)
        sp_data = self.training_dataset.spec_data
        sp_phases = self.model.get_restframe_phases(sp_data)
        lambda_c = lc_data.wavelength / (1. + lc_data.z)

        # too bad we need to call the model again here...
        # let's see whether we could cache this call
        flux = self.model(p, jac=False)

        J_lc = self.basis.eval(lambda_c, lc_phases)
        dsig = J_lc @ pars['gamma'].full
        sigma_lc = dsig * lc_data.flux
        # sigma_sp = np.zeros_like(sp_phases)
        # sig = np.hstack((sigma_lc, sigma_sp))
        var = sigma_lc**2

        if jacobian is not None:
            return var, None

            # n_pars, n_data = len(self.model.pars.free) + len(self.pars.free), len(sig)
            # v, jacobian = self.model(beta, jac=True)

            # i_dp, j_dp, jacobian_pars, i_dg0, j_dg0, jacobian_gamma = [], [], [], [], [], []

            # # dtmax
            # idx = self.model.pars['tmax'].indexof(sp['sn_id'])
            # jacobian_wrt_phase, jacobian_wrt_wavelength = self.basis.gradient(wl_sp,
            #                                                                   phase_sp + self.model.delta_phase)
            # if len(jacobian_pars) != 0.:
            #     jacobian_pars[idx] -= 2 * (jacobian_wrt_phase * self.pars['gamma'].full) / \
            #                           (1+sp['ZHelio']) * v[sp['i']] * sigma_spectra
            # # dg
            # i_dg0 = spline_sp.row.copy()
            # j_dg0 = self.pars['gamma'].indexof(spline_sp.col) + len(self.model.pars.free)
            # jacobian_gamma = 2 * v[sp['i']][i_dg0]**2 * spline_sp.data * ss_sp[i_dg0]
            # if debug_mode:
            #     return i_dp, j_dp, jacobian_pars, i_dg0, j_dg0, jacobian_gamma, v, sp, spline_sp, ss_sp

            # v_lc = v[lc['i']]
            # jacobian_lightcurve = jacobian.tocsc()[lc['i']].tocoo()

            # i_dp_lc = jacobian_lightcurve.row.copy()
            # j_dp_lc = jacobian_lightcurve.col.copy()
            # jacobian_pars_lightcurve = 2 * jacobian_lightcurve.data * (sigma_lightcurve * ss_lc)[i_dp_lc]

            # # dtmax
            # if self.model.pars['tmax'].indexof().sum() != len(self.model.pars['tmax'].full) * -1:
            #     jacobian_wrt_phase, jacobian_wrt_wavelength = self.basis.gradient(lambda_c,
            #                                                                             phase_lc +
            #                                                                             self.model.delta_phase)
            #     idx_lc = np.where(np.in1d(jacobian_lightcurve.col, self.model.pars['tmax'].indexof()))
            #     if len(jacobian_pars_lightcurve) != 0.:
            #         jacobian_pars_lightcurve[idx_lc] -= 2 * (jacobian_wrt_phase * self.pars['gamma'].full) / (1+lc['ZHelio']) * v_lc * sigma_lightcurve

            # # dg
            # i_dg0_lc = spline_lc.row.copy()
            # j_dg0_lc = self.pars['gamma'].indexof(spline_lc.col) + len(self.model.pars.free)
            # jacobian_gamma_lightcurve = 2 * v_lc[i_dg0_lc] ** 2 * spline_lc.data * ss_lc[i_dg0_lc]

            # if len(i_dg0) != 0:
            #     i_dg0 += len(lc)
            # if len(i_dp) != 0:
            #     i_dp += len(lc)

            # i = np.hstack((i_dp, i_dg0, i_dp_lc, i_dg0_lc))
            # j = np.hstack((j_dp, j_dg0, j_dp_lc, j_dg0_lc))
            # val = np.hstack((jacobian_pars, jacobian_gamma, jacobian_pars_lightcurve, jacobian_gamma_lightcurve))

            # idx_zeros = np.where((j != -1))[0]
            # jacobian_var = coo_matrix((val[idx_zeros], (i[idx_zeros], j[idx_zeros])), shape=(n_data, n_pars))

            # return var, jacobian_var.T

        return var



#!/usr/bin/env python

import os
import sys
from pathlib import Path
import logging
import pickle
from argparse import ArgumentParser

import matplotlib as mpl
import numpy as np
import numpy.ma as ma
import pylab as pl
import pandas

from lemaitre import bandpasses
from nacl import TrainingDataset, LogLikelihood, Minimizer
from nacl.models import salt2
from nacl.models.salt2 import constraints
from nacl.models.salt2.regularizations import NaClSplineRegularization
from nacl.models.salt2.variancemodels import SimpleErrorSnake, LocalErrorSnake

from saunerie.robuststat import mad


_band_ordering = {'u': 0, 'U': 0, 'g': 1, 'B': 2, 'r': 3, 'r2': 3,
                  'R': 4, 'i': 5, 'i2': 5, 'I': 6, 'z': 7, 'Y': 8}

def _get_band_order(band):
    b = band.split('::')[-1]
    return _band_ordering.get(b, 100)


def order_bands(bands):
    l = bands.sort(key=_get_band_order)
    return l


def clone(tds, sn, phase_range=(-25, 51., 1.)):
    """
    """
    # sn metadata
    sn_data = tds.sn_data.nt[tds.sn_data.sn == sn]
    
    # lightcurve data
    if tds.lc_data is not None:
        lc_data = tds.lc_data.nt[tds.lc_data.sn == sn]

        tmax = float(sn_data.tmax)
        z = float(sn_data.z)
        phase = np.arange(*phase_range)
        mjd = phase * (1.+z) + tmax
        N = len(mjd)
        l = []
        for b in np.unique(lc_data.band):
            d = lc_data[lc_data.band == b]
            Z = np.full(len(mjd), d[0])
            Z['mjd'] = mjd
            Z['flux'] = 0.
            Z['fluxerr'] = 0.
            Z['valid'] = 1
            l.append(Z)
        lc_data = np.rec.array(np.hstack(l))
    else:
        lc_data = None
        
    if tds.spec_data is not None:
        spec_data = tds.spec_data.nt[tds.spec_data.sn == sn]
    else:
        spec_data = None
    if tds.spectrophotometric_data is not None:
        spectrophotometric_data = tds.spectrophotometric_data.nt[tds.spectrophotometric_data.sn == sn]
    else:
          spectrophotometric_data = None
          
    
    tds = TrainingDataset(sn_data, lc_data=lc_data,
                          spec_data=None,
                          spectrophotometric_data=None,
                          filterlib=tds.filterlib)
    
    return tds


def plot_lc(sn, tds, pars=None, model=None, phase=False, v=None, norm=1.): # was norm=8.5E-13
    """
    """
    sel = tds.lc_data.sn == sn
    bands = np.unique(tds.lc_data.band[sel]).tolist()
    bands.sort(key=_get_band_order)
    nrows = len(bands)

    sn_idx = tds.sn_data.sn == sn
    tmax = float(tds.sn_data.tmax[sn_idx])
    z = float(tds.sn_data.z[sn_idx])
    sn_index = int(tds.sn_data.sn_index[sn_idx])

    if pars is not None:
        t = clone(tds, sn)
        m = salt2.SALT2Like(t)
        p = m.init_pars()
        for nm in ['M0', 'M1', 'CL']:
            p[nm].full[:] = pars[nm].full[:]
        for nm in ['X0', 'X1', 'col', 'tmax']:
            p[nm].full[0] = pars[nm].full[sn_index]
        v_pars = m(p)
    # else:
    #     v = v[sel]

    fig, axes = pl.subplots(nrows=nrows, ncols=1,
                            figsize=(8,12),
                            sharex=True)
    for i,b in enumerate(bands):
        idx = sel & (tds.lc_data.band == b)
        wl = tds.lc_data.wavelength[idx].mean()
        color = pl.cm.jet(int((wl-2000)/(9000.-2000.) * 256))

        if phase:
            xx = (tds.lc_data.mjd[idx] - tmax) / (1. + z)
            axes[i].axvline(0., ls=':')
        else:
            xx = tds.lc_data.mjd[idx]
            axes[i].axvline(tmax, ls=':')

        axes[i].errorbar(xx, tds.lc_data.flux[idx],
                         yerr=tds.lc_data.fluxerr[idx],
                         ls='', marker='.', color=color)

        if v is not None:
            vv = v[:len(tds.lc_data)]
            axes[i].plot(xx, norm*vv[idx], color=color, ls='', marker='+')

        if pars is not None:
            iii = t.lc_data.band == b
            if phase:
                xx = (t.lc_data.mjd[iii] - tmax) / (1. + z)
            else:
                xx = t.lc_data.mjd[iii]
            axes[i].plot(xx, norm * v_pars[iii], ls=':', color=color)

        axes[i].set_ylabel(b)

    axes[-1].set_xlabel('mjd')
    pl.subplots_adjust(hspace=0.05)
    fig.suptitle(f'{sn} @ z={z:4.3}')


def recalib_spectrum(spec, tds, v):
    """
    """
    sel = tds.spec_data.spec == spec
    vv = v[len(tds.lc_data):]

    fig, axes = pl.subplots(nrows=3, ncols=1, figsize=(8, 10),
                            sharex=True)
    axes[0].errorbar(tds.spec_data.wavelength[sel], tds.spec_data.flux[sel],
                     yerr=tds.spec_data.fluxerr[sel], ls='', marker='.')
    axes[0].plot(tds.spec_data.wavelength[sel], vv[sel], marker='.', color='r')

    r = tds.spec_data.flux[sel]/vv[sel]
    axes[1].errorbar(tds.spec_data.wavelength[sel], r,
                     yerr=tds.spec_data.fluxerr[sel], ls='', marker='.')

    recal = np.median(r)
    axes[2].errorbar(tds.spec_data.wavelength[sel], tds.spec_data.flux[sel],
                     yerr=tds.spec_data.fluxerr[sel], ls='', marker='.')
    axes[2].plot(tds.spec_data.wavelength[sel], vv[sel] * recal, marker='.', color='r')

    return recal


def recalib_spectra(tds, v):
    """
    """
    vv = v[len(tds.lc_data):]
    ii = np.arange(len(vv))
    idx = np.abs(tds.spec_data.fluxerr / tds.spec_data.flux) < 1.
    zero_fluxerr = np.abs(tds.spec_data.fluxerr) <= 0.
    logging.info(f'ignoring {zero_fluxerr.sum()} points with buggy flux errs')
    idx &= ~zero_fluxerr
    neg_flux = tds.spec_data.flux <= 0.
    idx &= ~neg_flux


    fig, axes = pl.subplots(nrows=3, ncols=1, figsize=(8, 10),
                            sharex=True)
    axes[0].errorbar(ii[idx], tds.spec_data.flux[idx],
                     yerr=tds.spec_data.fluxerr[idx], ls='', marker='.')
    axes[0].plot(ii[idx], vv[idx], marker='.', color='r')

    r = np.abs(tds.spec_data.flux / vv)
    r_err = tds.spec_data.fluxerr / np.abs(vv)
    axes[1].plot(ii[idx], r[idx],
                 ls='', marker='.')

    w = 1. / r_err**2
    n = np.bincount(tds.spec_data.spec_index[idx], weights=(w*r)[idx])
    d = np.bincount(tds.spec_data.spec_index[idx], weights=w[idx])
    print(n[1700], d[1700])
    recal = n / d

    v_recal = vv * recal[tds.spec_data.spec_index]

    #axes[2].errorbar(ii[idx], tds.spec_data.flux[idx],
    #                 yerr=tds.spec_data.fluxerr[idx], ls='', marker='.')
    axes[2].plot(ii[idx], tds.spec_data.flux[idx],
                 ls='', marker='.')
    axes[2].plot(ii[idx], v_recal[idx], ls='', marker='.', color='r')

    return recal


def plot_spec(spec, tds, v=None, norm=8.5E-13):
    """
    """
    sel = tds.spec_data.spec == spec
    vv = v[len(tds.lc_data):]

    fig, axes = pl.subplots(nrows=1, ncols=1,
                            figsize=(8,8),
                            sharex=True)

    axes.errorbar(tds.spec_data.wavelength[sel], tds.spec_data.flux[sel],
                     yerr=tds.spec_data.fluxerr[sel],
                     ls='', marker='.', color='blue')

    if v is not None:
        axes.plot(tds.spec_data.wavelength[sel], vv[sel], 'r+:')

    axes.set_ylabel('flux')
    axes.set_xlabel('mjd')
    pl.subplots_adjust(hspace=0.05)
    # xfig.suptitle(f'{sn} @ z={z:4.3}')


class TrainingChi2:

    def __init__(self, tds, pars, model):
        self.tds = tds
        self.pars = pars
        self.model = model
        self._compute()

    def _compute(self):
        """Compute SN/lc/spec partial chi2's
        """
        self.v = self.model(self.pars)
        self.res = self.tds.get_all_fluxes() - self.v
        self.yerr = self.tds.get_all_fluxerr()
        self.wres = self.res / self.yerr

        self.chi2 = (self.wres**2).sum()

        # photometry
        ml = len(tds.sn_data.sn_set)
        self.phot_wres = self.wres[:len(self.tds.lc_data)]
        self.phot_chi2 = (self.phot_wres**2).sum()
        self.phot_chi2_per_sn = np.bincount(self.tds.lc_data.sn_index, weights=self.phot_wres**2, minlength=ml)
        self.phot_chi2_per_sn_count = np.bincount(self.tds.lc_data.sn_index, minlength=ml)
        self.phot_chi2_per_lc = np.bincount(self.tds.lc_data.lc_index, weights=self.phot_wres**2, minlength=ml)
        self.phot_chi2_per_lc_count = np.bincount(self.tds.lc_data.lc_index, minlength=ml)

        # spectroscopy
        # spec_wres = wres[len(tds.lc_data):len(tds.lc_data) + len(tds.spec_data)]
        # spec_chi2 =

        self.r_phot_chi2_per_sn = self.phot_chi2_per_sn / self.phot_chi2_per_sn_count
        self.r_phot_chi2_per_lc = self.phot_chi2_per_lc / self.phot_chi2_per_lc_count

    def get_sorted_list(self, sort_by='sn'):
        """
        """
        if sort_by == 'sn':
            # i = np.argsort(self.r_phot_chi2_per_sn)[::-1]
            d = pandas.DataFrame({'sn': self.tds.sn_data.sn,
                                  'sn_index': self.tds.sn_data.sn_index,
                                  'z': self.tds.sn_data.z,
                                  'rchi2': self.r_phot_chi2_per_sn[self.tds.sn_data.sn_index]})
            d = d.sort_values(by='rchi2', ascending=False)

        elif sort_by == 'lc':
            # i = np.argsort(self.r_phot_chi2_per_lc)
            d = pandas.DataFrame({'lc_index': self.tds.lc_db.lc_index,
                                  'sn': self.tds.sn_data.sn_set[self.tds.lc_db.sn_index],
                                  'band': self.tds.lc_db.band,
                                  'rchi2': self.r_phot_chi2_per_lc[self.tds.lc_db.lc_index]})
            d = d.sort_values(by='rchi2', ascending=False)

        return d


def clean_dataset(tds, phase_range=(-25., +80.)):
    """
    """
    # clean ZTF19adcecwu
    idx = (tds.lc_data.sn == 'ZTF19adcecwu') & (tds.lc_data.mjd > 58840) & (tds.lc_data.flux<10000.)
    logging.info(f'removing {idx.sum()} outliers identified on ZTF19adcecwu g- and r- DR2 lightcurves')
    tds.lc_data.valid[idx] = 0

    # phase range (photometric data)
    tmax = np.zeros(len(tds.sn_data))
    tmax[tds.sn_data.sn_index] = tds.sn_data.tmax
    phot_tmax = tmax[tds.lc_data.sn_index]
    phase = (tds.lc_data.mjd - phot_tmax) / (1. + tds.lc_data.z)
    print(phase)
    print(phase_range)
    idx = (phase<phase_range[0]) | (phase>phase_range[1])
    logging.info(f'removing {idx.sum()} photometric points outside phase range')
    tds.lc_data.valid[idx] = 0

    # phase range (spectra)
    spec_tmax = tmax[tds.spec_data.sn_index]
    phase = (tds.spec_data.mjd - spec_tmax) / (1. + tds.spec_data.z)
    idx = (phase<phase_range[0]) | (phase>phase_range[1])
    logging.info(f'removing {idx.sum()} spectroscopic points outside phase range')
    tds.spec_data.valid[idx] = 0

    # points of the edge of the wavelength basis
    # no: this version does not what we want to do
    # i_basis_max = tds.spec_data.i_basis.max()
    # idx = (tds.spec_data.i_basis < 3) | (tds.spec_data.i_basis >= (i_basis_max-3))
    #
    # we need to determine, for each spectrum, the i_basis_min and i_basis_max
    b_bins = np.arange(-0.5, tds.spec_data.i_basis.max() + 1.5, 1)
    s_bins = np.arange(-0.5, tds.spec_data.spec_index.max() + 1.5, 1)
    h, _, _ = np.histogram2d(tds.spec_data.i_basis, tds.spec_data.spec_index, bins=(b_bins, s_bins))
    bb = np.arange(0, tds.spec_data.i_basis.max() + 1, 1)
    ss = np.arange(0, tds.spec_data.spec_index.max() + 1, 1)
    bb, ss = np.meshgrid(bb, ss)
    u = bb * h.T
    mb = ma.masked_array(u, mask=u==0.)
    b_min, b_max = np.array(np.min(mb, axis=1)), np.array(np.max(mb, axis=1))
    to_kill  = np.array(b_min).astype(int)[tds.spec_data.spec_index] == tds.spec_data.i_basis
    to_kill |= np.array(b_max).astype(int)[tds.spec_data.spec_index] == tds.spec_data.i_basis

    logging.info(f'removing {to_kill.sum()} spectroscopic points outside wavelength range')
    tds.spec_data.valid[to_kill] = 0

    # remove the spectral points with an uncertainty larger than 1.E6


    # clean all SNe below 0.02
    #idx = (tds.lc_data.z < 0.01)
    #logging.info(f'removing {idx.sum()} very low redshift SNe')
    #tds.lc_data.valid[idx] = 0

    tds.compress()


def plot_all_spectra(tds):
    """
    """
    b_bins = np.arange(-0.5, tds.spec_data.i_basis.max()+1.5, 1.)
    s_bins = np.arange(-0.5, tds.spec_data.spec_index.max()+1.5, 1.)
    ttt = tds.spec_db.copy()
    u = ttt.argsort(order='z')
    k = dict([(u[i], i) for i in range(len(u))])
    pl.hist2d(tds.spec_data.i_basis, np.array([k[i] for i in tds.spec_data.spec_index]),
              weights=tds.spec_data.z,
              norm=mpl.colors.LogNorm(),
              bins=(b_bins, s_bins))
    pl.colorbar()
    pl.xlabel('i_basis')
    pl.ylabel('spectrum (ordered by redshift)')

def main(lc_only=False, spec_only=False,
         model_phase_range=(-20., +50.),
         data_phase_range=(-20., +40.)):
    # trainin dataset
    fl = bandpasses.get_filterlib()
    tds = TrainingDataset.read_parquet('lemaitre.compressed.parquet',
                                       filterlib=fl)
    clean_dataset(tds, phase_range=data_phase_range)

    if lc_only:
        tds = TrainingDataset(sne=tds.sn_data.nt,
                              lc_data=tds.lc_data.nt,
                              basis=tds.basis,
                              filterlib=fl)
    elif spec_only:
        tds = TrainingDataset(sne=tds.sn_data.nt,
                              spec_data=tds.spec_data.nt,
                              basis=tds.basis,
                              filterlib=fl)

    # model 
    model = salt2.SALT2Like(tds,
                            # phase_range=model_phase_range,
                            wl_grid=tds.basis.grid,
                            spectrum_recal_degree=3)
    
    # likelihood 
    ll = LogLikelihood(model)
    pars = model.init_pars()


    # minimizer
    mn = Minimizer(ll)
    
    # preliminary control plots : how does the model compare with the data ?
    v = model(pars)

    # recalibrate the spectra
    r = recalib_spectra(tds, v)
    if not lc_only:
        ll.pars['SpectrumRecalibration'].full[3::4] = np.log(r)
        pars['SpectrumRecalibration'].full[3::4] = np.log(r)

    return fl, tds, model, ll, pars, mn, v


def compare_fits(tds, pars):
    """
    """
    z = np.zeros(len(tds.sn_data))
    x0 = np.zeros(len(tds.sn_data))
    x1 = np.zeros(len(tds.sn_data))
    col = np.zeros(len(tds.sn_data))
    tmax = np.zeros(len(tds.sn_data))

    z[tds.sn_data.sn_index] = tds.sn_data.z
    x0[tds.sn_data.sn_index] = tds.sn_data.x0
    x1[tds.sn_data.sn_index] = tds.sn_data.x1
    col[tds.sn_data.sn_index] = tds.sn_data.col
    tmax[tds.sn_data.sn_index] = tds.sn_data.tmax

    pl.figure()
    pl.plot(z, x0/ll.pars['X0'].full, 'b.')
    pl.title('X0')
    pl.xlabel('z')
    pl.ylabel('$\Delta X0$')

    pl.figure()
    pl.plot(z, (x1-ll.pars['X1'].full), 'b.')
    pl.title('X1')
    pl.xlabel('z')
    pl.ylabel('$\Delta X1$')

    pl.figure()
    pl.plot(z, (col-ll.pars['col'].full), 'b.')
    pl.title('col')
    pl.xlabel('z')
    pl.ylabel('$\Delta col$')

    pl.figure()
    pl.plot(z, (tmax-ll.pars['tmax'].full), 'b.')
    pl.title('tmax')
    pl.xlabel('z')
    pl.ylabel('$\Delta tmax$')



def load_tds(fn='lemaitre.compressed.parquet',
             data_phase_range=(-20., 40.)):
    """load a training dataset
    """
    fl = bandpasses.get_filterlib()
    tds = TrainingDataset.read_parquet('lemaitre.compressed.parquet',
                                       filterlib=fl)
    clean_dataset(tds, phase_range=data_phase_range)


    # if lc_only:
    #     tds = TrainingDataset(sne=tds.sn_data.nt,
    #                           lc_data=tds.lc_data.nt,
    #                           basis=tds.basis,
    #                           filterlib=fl)
    # elif spec_only:
    #     tds = TrainingDataset(sne=tds.sn_data.nt,
    #                           spec_data=tds.spec_data.nt,
    #                           basis=tds.basis,
    #                           filterlib=fl)

    return tds


class FitRes:
    pass


def fit_local_pars(tds, model, res_init, variance_model=None, plot=False):
    """
    """
    # fl, tds, model, ll, pars, mn, v = main(lc_only=lc_only)
    ll = LogLikelihood(model, variance_model=variance_model)
    ll.pars.fix()
    ll.pars.release('X0')
    ll.pars.release('X1')
    ll.pars.release('col')
    ll.pars.release('tmax')
    ll.pars.release('SpectrumRecalibration')

    # recalibrate the spectra
    v = model(ll.pars)
    r = recalib_spectra(tds, v)
    ll.pars['SpectrumRecalibration'].full[3::4] = np.log(r)
    print(ll.pars['SpectrumRecalibration'].full[:20])

    minz = Minimizer(ll)
    p = minz.minimize_lm(ll.pars.free)
    if plot:
        minz.plot()

    res = FitRes()
    res.ll = ll
    res.minz = minz

    return res


class ModelPlotter:
    """A utility class to plot a smooth continuous (oversampled) photometric model
    """

    def __init__(self, model, sn_index, init_pars, variance_model=None):
        """Constructor - evaluate the model and error model
        """
        self.model = model
        self.tds = model.training_dataset
        self.init_pars = init_pars
        self.variance_model = variance_model
        self._init_local_pars(init_pars, sn_index)

        # evaluate model and variance model
        self.v = self.model(self.pars)
        if variance_model is not None:
            self.v_var = self.variance_model(self.pars)
        else:
            self.v_var = None

    def _init_local_pars(self, init_pars, sn_index):
        """initialize small parameter vector from many-sne one
        """
        if self.variance_model is not None:
            ll = LogLikelihood(model, variance_model=variance_model)
            self.pars = ll.pars.copy()
        else:
            self.pars = self.model.init_pars()

        for nm in ['M0', 'M1', 'CL']:
            self.pars[nm].full[:] = init_pars[nm].full[:]
        for nm in ['X0', 'X1', 'col', 'tmax']:
            self.pars[nm].full[0] = init_pars[nm].full[sn_index]
        if 'gamma' in init_pars._struct.slices:
            self.pars['gamma'].full[:] = init_pars['gamma'].full[:]
        else:
            self.pars['gamma'].full[:] = 0.

        return self.pars

    def plot(self, band, ax=None, **kwargs):
        """plotter - plot the model and error model if needed
        """
        if not ax:
            ax = pl.gca()

        phase = kwargs.get('phase', False)
        color = kwargs.get('color', 'b')
        ls = kwargs.get('ls', ':')
        label = kwargs.get('label', '')
        xlabel= kwargs.get('xlabel', None)
        ylabel= kwargs.get('ylabel', None)
        title = kwargs.get('title', None)
        alpha = kwargs.get('alpha', 0.25)
        legend = kwargs.get('legend', None)

        tmax = float(self.tds.sn_data.tmax[0])
        z = float(self.tds.sn_data.z[0])

        idx = self.tds.lc_data.band == band
        if phase:
            xx = (self.tds.lc_data.mjd - tmax) / (1. + z)
            ax.axvline(0., ls=':')
        else:
            xx = self.tds.lc_data.mjd
            ax.axvline(tmax, ls=':')

        ax.plot(xx[idx], self.v[idx], color=color, ls=ls, label=label)
        if self.variance_model is not None:
            vm = (self.v[idx] - np.sqrt(self.v_var[idx]))
            vp = (self.v[idx] + np.sqrt(self.v_var[idx]))
            ax.fill_between(xx[idx], vm, vp, alpha=0.5, color=color)

        if xlabel is not None:
            ax.set_xlabel(xlabel)
        if ylabel is not None:
            ax.set_ylabel(ylabel)
        if title is not None:
            ax.set_title(title)
        if legend:
            ax.legend(loc=legend)

class Trainer:

    def __init__(self, tds, model, variance_model=None):
        self.tds = tds
        self.model = model
        self.variance_model = variance_model
        self.log = []

        # recalibrate the spectra
        ll = LogLikelihood(self.model)
        self.v_init = self.model(ll.pars)
        self.spec_recal_init = recalib_spectra(self.tds, self.v_init)
        # ll.pars['SpectrumRecalibration'].full[3::4] = np.log(r)

    def fit(self, fix=None, p_init=None,
            p_init_blocks=None,
            fit_variance_model=False,
            force_spec_recalibration=False,
            max_iter=100):
        """
        """
        if fit_variance_model:
            ll = LogLikelihood(self.model, variance_model=self.variance_model)
        else:
            ll = LogLikelihood(self.model)

        if force_spec_recalibration:
            ll.pars['SpectrumRecalibration'].full[3::4] = np.log(self.spec_recal_init)

        if fix:
            for block_name in fix:
                ll.pars[block_name].fix()

        if p_init:
            for block_name in p_init._struct.slices:
                if block_name not in ll.pars._struct.slices:
                    continue
                ll.pars[block_name].full[:] = p_init[block_name].full[:]

        if p_init_blocks:
            for block_name in p_init_blocks:
                if block_name not in ll.pars._struct.slices:
                    continue
                ll.pars[block_name].full[:] = p_init_blocks[block_name]

        print(ll.pars)

        # fit
        minz = Minimizer(ll)
        p = minz.minimize_lm(ll.pars.free, max_iter=max_iter)

        res = FitRes()
        res.pars = ll.pars
        res.ll = ll
        res.minz = minz
        res.v = model(res.ll.pars, jac=0)
        res.v_var = None
        if self.variance_model is not None and fit_variance_model:
            res.v_var = self.variance_model(res.ll.pars, jac=0)

        return res

    def train(self):
        """Full model training
        """
        # initialization: PETS sncosmo/salt2 fits
        f0 = FitRes()
        f0.ll = LogLikelihood(self.model)
        f0.pars = f0.ll.pars
        f0.v = self.model(f0.pars)
        f0.var_v = None
        self.log.append(f0)

        # initialization : fit the light curves and spectra
        # no error model
        f1 = self.fit(fix=['M0', 'M1', 'CL'],
                      fit_variance_model=False,
                      force_spec_recalibration=True,
                      max_iter=10)
        self.log.append(f1)

        # fit an error model, all other pars fixed
        f2 = self.fit(fix=['M0', 'M1', 'CL', 'X0', 'X1', 'col', 'tmax', 'SpectrumRecalibration'],
                      fit_variance_model=True,
                      p_init=f1.ll.pars,
                      p_init_blocks={'gamma': -5.},
                      max_iter=10)
        self.log.append(f2)

        # refit the light curves and spectra,
        # with an error model
        f3 = self.fit(fix=['M0', 'M1', 'CL'],
                      fit_variance_model=True,
                      p_init=f2.ll.pars,
                      max_iter=10)
        self.log.append(f3)

        # train
        #f4 = self.fit(variance_model=True,
        #              p_init=f3.ll.pars)
        #self.log.append(f4)

    def _get_models_to_plot(self, numfit, pars):
        """
        """
        fits = []

        if numfit is not None:
            try:
                iter(numfit)
                fits.extend(self.log[i] for i in numfit)
            except:
                fits.append(self.log[numfit])

        if pars is not None:
            try:
                iter(pars)
                for p in pars:
                    ff = FitRes()
                    ff.pars = pars
                    ff.v = self.model(pars)
                    if self.variance_model:
                        ff.var_v = self.variance_model(pars)
                    fits.append(ff)
            except:
                ff = FitRes()
                ff.pars = pars
                ff.v = self.model(pars)
                if self.variance_model:
                    ff.var_v = self.variance_model(pars)
                fits.append(ff)

        return fits


    def plot_lc(self, sn, numfit=None, pars=None,
                phase=False, norm=1.,
                plot_variance=False):
        """plot light curve data + models

        Parameters
        ----------
        sn : (int | str)
          the supernova to plot
        numfit : int, optional
          which fit to plot (drawn from logs)
        pars : FitParameters, optional
          plot an alternate fit, not from self.log
        phase : bool, default False
          whether to plot the the light curve as a function of the phase
        plot_variance : bool, default=False
          whether to plot the error model, if available

        Returns
        -------
        None
        """
        sel = self.tds.lc_data.sn == sn
        bands = np.unique(self.tds.lc_data.band[sel]).tolist()
        bands.sort(key=_get_band_order)
        nrows = len(bands)

        sn_idx = self.tds.sn_data.sn == sn
        tmax = float(self.tds.sn_data.tmax[sn_idx])
        z = float(self.tds.sn_data.z[sn_idx])
        sn_index = int(self.tds.sn_data.sn_index[sn_idx])

        # clone the dataset
        t = clone(tds, sn)
        to_plot = self._get_models_to_plot(numfit, pars)
        m = salt2.SALT2Like(t)
        if self.variance_model is not None:
            vm = self.variance_model.__class__(m)
        else:
            vm = None
        plotters = [ModelPlotter(m, sn_index, tp.ll.pars, variance_model=vm) \
                    for tp in to_plot]

        # create a model
        # m = salt2.SALT2Like(t)
        # if self.variance_model is not None:
        #     vm = self.variance_model.__class__(m)
        # else:
        #     vm = None
        # ll = LogLikelihood(m, variance_model=vm)
        # p = ll.init_pars()

        # which models do we want to plot ?
        # to_plot = self._eval(numfit, pars, sn_index)

        # for tp in to_plot:
        #     for nm in ['M0', 'M1', 'CL']:
        #         p[nm].full[:] = tp.pars[nm].full[:]
        #     for nm in ['X0', 'X1', 'col', 'tmax']:
        #         p[nm].full[0] = tp.pars[nm].full[sn_index]
        #     if 'gamma' in tp.pars._struct.slices:
        #         p['gamma'].full[:] = tp.pars['gamma'].full[:]
        #     else:
        #         p['gamma'].full[:] = 0.
        #     v_pars = m(p)
        #     v_pars_err = vm(p) if vm is not None else None

        # plots
        fig, axes = pl.subplots(nrows=nrows, ncols=1,
                                figsize=(8,12),
                                sharex=True)
        for i,b in enumerate(bands):
            idx = sel & (tds.lc_data.band == b)
            wl = tds.lc_data.wavelength[idx].mean()
            color = pl.cm.jet(int((wl-2000)/(9000.-2000.) * 256))

            if phase:
                xx = (tds.lc_data.mjd[idx] - tmax) / (1. + z)
                axes[i].axvline(0., ls=':')
            else:
                xx = tds.lc_data.mjd[idx]
                axes[i].axvline(tmax, ls=':')

            axes[i].errorbar(xx, tds.lc_data.flux[idx],
                             yerr=tds.lc_data.fluxerr[idx],
                             ls='', marker='.', color=color)

            # plot the model
            # iii = t.lc_data.band == b
            # if phase:
            #     xx = (t.lc_data.mjd[iii] - tmax) / (1. + z)
            # else:
            #     xx = t.lc_data.mjd[iii]
            # axes[i].plot(xx, norm * v_pars[iii], ls=':', color=color)

            # # plot the variance model
            # vm = norm * (v_pars[iii] - np.sqrt(v_pars_err[iii]))
            # vp = norm * (v_pars[iii] + np.sqrt(v_pars_err[iii]))
            # axes[i].fill_between(xx, vm, vp,
            #                      color=color,
            #                      alpha=0.1)
            for p in plotters:
                p.plot(b, ax=axes[i],
                       color=color,
                       phase=phase,
                       ylabel=b)
            if phase:
                axes[-1].set_xlabel('mjd [days]')
            else:
                axes[-1].set_xlabel('phase [restframe days]')
            pl.subplots_adjust(hspace=0.05)
            fig.suptitle(f'{sn} @ z={z:4.3}')

    def plot_spec(self, spec, numfit=None, pars=None, alpha=0.5):
        """
        """
        sel = tds.spec_data.spec == spec

        fig, axes = pl.subplots(nrows=1, ncols=1,
                                figsize=(8,8),
                                sharex=True)

        axes.errorbar(tds.spec_data.wavelength[sel], tds.spec_data.flux[sel],
                      yerr=tds.spec_data.fluxerr[sel],
                      ls='', marker='.', color='blue')

        to_plot = self._get_models_to_plot(numfit, pars)
        for tp in to_plot:
            vv = tp.v[len(tds.lc_data):][sel]
            axes.plot(tds.spec_data.wavelength[sel], vv, 'r+:')
            if tp.v_var is not None:
                vv_var = tp.v_var[len(tds.lc_data):][sel]
                vmin = vv - np.sqrt(vv_var)
                vmax = vv + np.sqrt(vv_var)
                axes.fill_between(tds.spec_data.wavelength[sel],
                                  vmin, vmax,
                                  color='r', alpha=alpha)

        axes.set_ylabel('flux')
        axes.set_xlabel('mjd')

        sn = np.unique(tds.spec_data.sn[sel])
        assert len(sn) == 1
        sn = sn[0]
        z = np.unique(tds.spec_data.z[sel])
        assert len(z) == 1
        z = z[0]
        axes.set_title(f'{sn} @ z={z:4.3}')
        pl.subplots_adjust(hspace=0.05)

    def cut_lc_outliers(self, pars, nsig=10):
        """
        """
        v = self.model(pars)
        if self.variance_model is not None and 'gamma' in pars._struct.slices:
            var_v = self.variance_model(pars)
        else:
            var_v = None

        vv = v[:len(self.tds.lc_data)]
        if var_v is not None:
            var_vv = var_v[:len(self.tds.lc_data)]
        else:
            var_vv = None

        bands = np.unique(self.tds.lc_data.band)
        nbands = len(bands)

        for i,band in enumerate(bands):
            sel = self.tds.lc_data.band == band
            var = self.tds.lc_data.fluxerr**2
            if var_vv is not None:
                var += var_vv
            w = 1. / var
            res = vv - self.tds.lc_data.flux
            wres = np.sqrt(w) * res

            cut = (np.abs(wres) > nsig)
            cc = cut[sel]
            logging.info(f'cutting {cc.sum()}/{len(cc)}={100*cc.sum()/len(cc):4.3}% outliers in {band}')
            self.tds.lc_data.valid[sel] = (~cut)[sel]

            jj = (~cut)[sel]
            print(jj.sum(), self.tds.lc_data.valid[sel].sum(), len(self.tds.lc_data.valid[sel]))


    def plot_lc_training_residuals(self, pars, v=None, var_v=None,
                                   phases=False,
                                   weighted_residuals=False,
                                   hexbin=False):
        """global fit residuals
        """
        if not v:
            v = self.model(pars)
        if not var_v and self.variance_model is not None and 'gamma' in pars._struct.slices:
            var_v = self.variance_model(pars)

        vv = v[:len(tds.lc_data)]
        if var_v is not None:
            var_vv = var_v[:len(tds.lc_data)]
        else:
            var_vv = None

        tmax = pars['tmax'].full[self.tds.lc_data.sn_index]
        z = self.tds.lc_data.z

        bands = np.unique(tds.lc_data.band)
        nbands = len(bands)
        fig, axes = pl.subplots(figsize=(6,12),
                                nrows=nbands, ncols=2,
                                sharex=0, sharey=1,
                                gridspec_kw={'width_ratios': [3,1]})
        for i, band in enumerate(bands):
            sel = tds.lc_data.band == band
            wl = tds.lc_data.wavelength[sel].mean()
            color = pl.cm.jet(int((wl-2000)/(9000.-2000.) * 256))
            if phases:
                x = (self.tds.lc_data.mjd - tmax) / (1. + z)
            else:
                x = np.arange(len(sel))

            var = tds.lc_data.fluxerr**2
            if var_vv is not None:
                var += var_vv
            w = 1. / var
            res = vv - tds.lc_data.flux
            wres = np.sqrt(w) * res
            y = wres if weighted_residuals else res
            if hexbin:
                axes[i,0].hexbin(x[sel], y[sel],
                                 gridsize=(100, 500),
                                 extent=(-30, 50, -100, 100),
                                 mincnt=1)
            else:
                axes[i,0].plot(x[sel], y[sel],
                               # yerr=tds.lc_data.fluxerr[sel],
                               ls='', marker='.', color=color)
            axes[i,1].hist(y[sel], bins=500,
                           color=color,
                           density=True,
                           # orientation='vertical')
                           )
            print(band, mad(y[sel]))

            # if var_vv is not None:
            #    ii = np.argsort(x[sel])
            # axes[i].fill_between(x[sel][ii], -np.sqrt(var_vv[sel][ii]), np.sqrt(var_vv[sel][ii]),
            #                      alpha=0.25, color=color)
            axes[i,0].set_ylabel(band)
        axes[-1,0].set_xlabel('phases [days]' if phases else 'index')
        pl.subplots_adjust(hspace=0.01, wspace=0.01)

        return v, var_v


    def plot_spec_training_residuals(self, pars, v=None, var_v=None):
        """global fit residuals
        """
        if not v:
            v = self.model(pars)
        if not var_v and self.variance_model is not None and 'gamma' in pars._struct.slices:
            var_v = self.variance_model(pars)

        vv = v[len(self.tds.lc_data):]
        if var_v is not None:
            var_vv = var_v[len(self.tds.lc_data):]
        else:
            var_vv = None

        sel = self.tds.spec_data.valid

        tmax = pars['tmax'].full[self.tds.spec_data.sn_index]
        z = self.tds.spec_data.z
        phases = (self.tds.spec_data.mjd - tmax) / (1. + z)

        fig, axes = pl.subplots(nrows=1, ncols=1)
        # color = pl.cm.jet(int(phases/(50+25) * 256))
        x = self.tds.spec_data.wavelength
        ii = np.argsort(x)
        axes.errorbar(x[sel], vv[sel] - tds.spec_data.flux[sel],
                      yerr=tds.spec_data.fluxerr[sel],
                      ls='', marker=',', color='gray')
        #        if var_vv is not None:
        #            axes[i].fill_between(x[sel][ii], -np.sqrt(var_vv[sel][ii]), np.sqrt(var_vv[sel][ii]),
        #                                 alpha=0.25, color='gray')
        axes.set_xlabel('wavelength [$\AA$]')
        pl.subplots_adjust(hspace=0.05)

        return v, var_v


    def plot_lightcurves(self, pars, v=None, var_v=None, phases=False):
        """global fit residuals
        """
        if not v:
            v = self.model(pars)
        if not var_v and self.variance_model is not None and 'gamma' in pars._struct.slices:
            var_v = self.variance_model(pars)

        vv = v[:len(tds.lc_data)]
        if var_v is not None:
            var_vv = var_v[:len(tds.lc_data)]
        else:
            var_vv = None

        tmax = pars['tmax'].full[self.tds.lc_data.sn_index]
        z = self.tds.lc_data.z

        bands = np.unique(tds.lc_data.band)
        nbands = len(bands)
        fig, axes = pl.subplots(nrows=nbands, ncols=1, sharex=phases)
        for i, band in enumerate(bands):
            sel = tds.lc_data.band == band
            wl = tds.lc_data.wavelength[sel].mean()
            color = pl.cm.jet(int((wl-2000)/(9000.-2000.) * 256))
            if phases:
                x = (self.tds.lc_data.mjd - tmax) / (1. + z)
            else:
                x = self.tds.lc_data.mjd
            ii = np.argsort(x[sel])

            axes[i].errorbar(x[sel][ii], tds.lc_data.flux[sel][ii],
                             yerr=tds.lc_data.fluxerr[sel][ii],
                             ls='', marker=',', color=color)
            axes[i].plot(x[sel][ii], vv[sel][ii],
                         ls='', marker='+', color=color)

            if var_vv is not None:
                axes[i].fill_between(x[sel][ii], vv[sel]-np.sqrt(var_vv[sel][ii]), vv[sel]+np.sqrt(var_vv[sel][ii]),
                                     alpha=0.5, color=color)
            axes[i].set_ylabel(band)
        axes[-1].set_xlabel('phases [days]' if phases else 'index')
        pl.subplots_adjust(hspace=0.05)

        return v, var_v




    def save(self, fn):
        with open(fn, 'wb') as f:
            pickle.dump(self, f)

    @classmethod
    def load(cls, fn):
        with open(fn, 'rb') as f:
            return pickle.load(f)

    def plot(self, res):
        """
        """
        t = TrainingChi2(self.tds, res.ll.pars, res.ll.model)
        d = t.get_sorted_list(sort_by='sn')

        # plot the 5 worst SNe
        for nm in d[:5].sn:
            plot_lc(nm, self.tds, res.ll.pars, norm=1, phase=1)

        for spec in [222, 333, 444, 555, 1226, 1500, 1700]:
            plot_spec(spec, tds, res.v)




def fit_local_pars_and_error_model(tds, model, plot=False):
    """
    """
    # fl, tds, model, ll, pars, mn, v = main(lc_only=lc_only)
    error_snake = SimpleErrorSnake(model)
    ll = LogLikelihood(model, variance_model=error_snake)
    ll.pars.fix()
    ll.pars.release('X0')
    ll.pars.release('X1')
    ll.pars.release('col')
    ll.pars.release('tmax')
    ll.pars.release('SpectrumRecalibration')
    ll.pars.release('gamma')
    ll.pars['gamma'].full[:] = 0.05

    # recalibrate the spectra
    v = model(ll.pars)
    r = recalib_spectra(tds, v)
    ll.pars['SpectrumRecalibration'].full[3::4] = np.log(r)
    print(ll.pars['SpectrumRecalibration'].full[:20])

    minz = Minimizer(ll)
    p = minz.minimize_lm(ll.pars.free)
    if plot:
        minz.plot()

    res = FitRes()
    res.ll = ll
    res.minz = minz

    return res


def fit_error_model(tds, model, res_init, plot=False):
    """
    """
    # fl, tds, model, ll, pars, mn, v = main(lc_only=lc_only)
    error_snake = SimpleErrorSnake(model)
    ll = LogLikelihood(model, variance_model=error_snake)
    ll.pars.fix()
    ll.pars.release('gamma')
    ll.pars['gamma'] = 0.05

    for block_name in res_init.ll.pars._struct.slices:
        ll.pars[block_name].full[:] = res_init.ll.pars[block_name].full[:]

    minz = Minimizer(ll)
    p = minz.minimize_lm(ll.pars.free)
    if plot:
        minz.plot()

    res = FitRes()
    res.ll = ll
    res.minz = minz

    return res


def main():
    train = train = Trainer(tds, model, variance_model)
    train.train()
    train.save('training.pkl')

    # plot 5 worst and 5 best light curves and spectra
    # train.plot()


    # plot training residuals
    train.plot_lc_training_residuals()
    train.plot_spec_training_residuals()


def fit_model(tds, model, ll_init, plot=False):
    """
    """
    cons = constraints.salt2like_linear_constraints(model, mu=1.E6)
    # TODO: ll_init.pars ?  Chicken and egg problem !
    reg = NaClSplineRegularization(model, ll_init.pars, to_regularize=['M0', 'M1'], mu=1.)
    ll = LogLikelihood(model, cons=[cons], reg=[reg])

    ll.pars.full[:] = ll_init.pars[:].full
    ll.pars.fix()
    ll.pars.release('M0')
    ll.pars.release('M1')
    ll.pars.release('CL')

    minz = Minimizer(ll)
    p = minz.minimize_lm(ll.pars.free)
    if plot:
        minz.plot()

    res = FitRes()
    res.ll = ll
    res.minz = minz
    res.cons = cons
    res.reg = reg

    return res


if __name__ == '__main__':

    tds, model, variance_model = None, None, None
    ll_init = None

    parser = ArgumentParser(description='NaCl training on Lemaitre')
    parser.add_argument('--reload',
                        action='store_true',
                        help='force reloading tds and model')
    parser.add_argument('--fit-local-pars',
                        action='store_true',
                        help='fit the supernova parameters (fixed model)')
    parser.add_argument('--fit-local-pars-and-error-model',
                        action='store_true',
                        help='fit the supernova parameters with an error model (fixed model)')
    parser.add_argument('--fit-error-model',
                        action='store_true',
                        help='fit an error model (fixed local pars and model)')
    parser.add_argument('--fit-model',
                        action='store_true',
                        help='fit the model (fixed local pars)')
    parser.add_argument('--fit-model-and-error-model',
                        action='store_true',
                        help='fit the model along with an error model')
    # parser.add_argument('--fit-model-and-error-model',
    #                     action='store_true',
    #                     help='fit the model (fixed local pars) along with an error model')
    # parser.add_argument('--train',
    #                     action='store_true',
    #                     help='fit all parameters')

    args = parser.parse_args()

    if tds is None or model is None and not args.force_reload:
        tds = load_tds()
        tds.compress()
        model = salt2.SALT2Like(tds,
                                # phase_range=model_phase_range,
                                wl_grid=tds.basis.grid,
                                spectrum_recal_degree=3)
        variance_model = LocalErrorSnake(model)

#    if args.fit_local_pars:
#        res_init = fit_local_pars(tds, model)
#    #    elif args.fit_local_pars_and_error_model:
#    #        res_init = fit_local_pars_and_error_model(tds, model)

#    if args.fit_error_model:
#        res = fit_error_model(tds, model, res_init)

#    if args.fit_model:
#        if ll_init is None:
#            res_init = fit_local_pars(tds, model)
#        res = fit_model(tds, model, ll_init=res_init.ll)
#    elif args.fit_model_and_error_model:
#        res = fit_model_and_error_model(tds, model, ll_init=res_init.ll)


    # if sys.argv[1] == '2':
    #     cons = constraints.salt2like_linear_constraints(model, mu=1.E6) # was 1.E6
    #     reg = NaClSplineRegularization(model, pars, to_regularize=['M0', 'M1'], mu=1.) # was 1.E-6
    #     ll2 = LogLikelihood(model, reg=[reg])
    #     ll2.pars.full[:] = ll.pars[:].full
    #     ll2.pars.fix()
    #     ll2.pars.release('M0')
    #     ll2.pars.release('M1')
    #     ll2.pars.release('CL')
    #     mn2 = Minimizer(ll2)
    #     p = mn2.minimize_lm(ll2.pars.free)
    #     mn2.plot()

    # if sys.argv[1] == '3':
    #     ll3 = LogLikelihood(model, reg=[reg], cons=[cons]) # TODO cons make sure it is a list
    #     ll3.pars.full[:] = ll2.pars[:].full
    #     ll3.pars.release()
    #     mn3 = Minimizer(ll3)
    #     p = mn3.minimize_lm(ll3.pars.free)
    #     mn3.plot()


    #    ll.pars.release('M1')




"""SALT2 color-law
"""

import numpy as np


class ColorLaw:
    r"""Implementation of the SALT2 color law (version 1)

        The SALT2 color law describes the color diversity of SNeIa. In SALT 2.4, the
        color law is parametrized as follows:

        .. math::
            CL(\lambda;c) = 10^{0.4\ c \times P(\lambda_r)}

        where :math:`P(\lambda_r)` is a polynomial, such as :math:`P(\lambda_B) = 0`
        and :math:`P(\lambda_V) = 1`.

        This implies the following transformation :

        .. math::
            \lambda_r = \frac{\lambda-\lambda^B}{\lambda^V - \lambda^B}

        It is defined in the wavelength range :math:`2800 A < \lambda < 7000 A` and
        extrapolated with a linear function outside this interval.

        This class evaluates :math:`P(\lambda, \alpha_i)` and its derivatives
        :math:`\partial P/\partial \alpha_i` (which is the painful part).

    """
    WAVELENGTH = {"U": 3650.88, "B": 4302.57, "V": 5428.55, "R": 6418.01, "I": 7968.34}
    U_WAVELENGTH = 3650.88
    B_WAVELENGTH = 4302.57
    V_WAVELENGTH = 5428.55
    R_WAVELENGTH = 6418.01
    I_WAVELENGTH = 7968.34

    def __init__(self, wl_range=[2800, 7000]):
        r"""Constructor

        Parameters
        ----------
        wl_range : array_like
            nominal wavelength range for the polynom color law. Outside this
            range, the color law is extrapolated with deg-1 polynomials.

        """
        assert (wl_range[0] < wl_range[1])
        self.min_lambda, self.max_lambda = wl_range

        self.min_reduced_lambda = self.reduce(self.min_lambda)
        self.max_reduced_lambda = self.reduce(self.max_lambda)
        self.pars = np.zeros(5)

    def reduce(self, wl):
        r"""wavelength remapping

        defined as:

        .. math::
            \lambda_r = \frac{\lambda - \lambda_B}{\lambda_V - \lambda_B}

        Parameters
        ----------
        wl : array-like
            the input wavelengths in Angstroms

        Returns
        -------
        reduced wavelenghts : array of floats
        """
        B_WL, V_WL = self.WAVELENGTH["B"], self.WAVELENGTH["V"]
        return (wl-B_WL)/(V_WL-B_WL)

    def __call__(self, wl, p, jac=False):  # return_jacobian_as_coo_matrix=False):
        r"""Evaluate the polynomial part of the color law

        The full SALT2 color law is given by:

        .. math::
            CL(\lambda, c) = 10^{0.4\ c\ P(\lambda, \alpha_i)}

        with

       .. math::
           P(\lambda) = \left\{ \begin{split}
           P'(\bar\lambda_{UV}) \times (\bar\lambda-\bar\lambda_{UV}) + P(\bar\lambda_{UV})
           & \ \ \ \ \mathrm{if \lambda < \lambda_{UV}} \\
           \bar\lambda \times \left(\sum_{i=1}^4 \alpha_i \bar\lambda^i + 1 - \sum_{i=1}^4\alpha_i\right)
           & \ \ \ \ \mathrm{if \lambda_{UV} < \lambda < \lambda_{IR}} \\
           P'(\bar\lambda_{IR}) \times (\bar\lambda-\bar\lambda_{IR}) + P(\bar\lambda_{IR})
           & \ \ \ \ \mathrm{if \lambda > \lambda_{IR}} \\
           \end{split}\right.

       and :math:`\lambda_r` defined above.

       This function evaluates :math:`P(\lambda,\alpha_i)` along with its
       derivatives w.r.t. the :math:`\alpha_i`: :math:`\partial P/\partial
       \alpha_i`

       Parameters
       ----------
       wl : ndarray of float
           input wavelengths (in Angstroms)
       p : ndarray
           color law parameters (i.e. the :math:`(\alpha_i)_{1\leq i \leq 4}`)
       jac : bool (default=False)
           whether to return the jacobian matrix
       return_jacobian_as_coo_matrix : bool
           If jacobian needs to be scipy.sparse

       Returns
       -------
       cl : array-like of float
           the color law values
       jacobian : numpy.array or scipy.sparse.csr_matrix (return_jacobian_as_coo_matrix)
           the matrix of derivatives, if `jac` is True
        """
        self.pars[0:4] = p
        self.pars[4] = 1 - p.sum()
        d_pars = np.polyder(self.pars)

        # nominal range
        rwl = self.reduce(np.asarray(wl))
        r = np.polyval(self.pars, rwl) * rwl

        # uv side (if necessary)
        idx_uv = rwl < self.min_reduced_lambda
        has_uv_data = idx_uv.sum() > 0
        if has_uv_data:
            val = np.polyval(self.pars, self.min_reduced_lambda)
            d_val = np.polyval(d_pars, self.min_reduced_lambda)
            self.pars_uv = np.asarray([d_val * self.min_reduced_lambda + val,
                                       val * self.min_reduced_lambda])
            r[idx_uv] = np.polyval(self.pars_uv, rwl[idx_uv]-self.min_reduced_lambda)

        # ir side
        idx_ir = rwl > self.max_reduced_lambda
        has_ir_data = idx_ir.sum() > 0
        if has_ir_data:
            val = np.polyval(self.pars, self.max_reduced_lambda)
            d_val = np.polyval(d_pars, self.max_reduced_lambda)
            self.pars_ir = np.asarray([d_val * self.max_reduced_lambda + val,
                                       val * self.max_reduced_lambda])
            r[idx_ir] = np.polyval(self.pars_ir, rwl[idx_ir]-self.max_reduced_lambda)

        if not jac:
            return r, None

        # the jacobian is unfortunately a dense matrix
        # (maybe we should try using splines)
        #
        # In the nominal wavelength range, it has the form:
        #
        #   [l_1**5-l_1   l_1**4-l1    l_1**3-l1  l_1**2-l1]
        #   [  ...          ...           ...       ...    ]
        #   [l_N**5-l_1   l_N**4-l1    l_N**3-l1  l_N**2-l1]
        #
        v = np.vander(rwl, 6)[:, 0:-2]
        jacobian = (v.T-rwl).T

        #
        # and in the extrapolation range, it has the form:
        #
        # J_ik = [dCL'/da_k(rwl-rwl_uv) + dCL/da_k]
        #
        # Granted, it's not very readable. But it should be fast enough
        if has_uv_data:
            l_uv = self.min_reduced_lambda
            j1_luv = (np.vander([l_uv], 6)[:, 0:-2].T - np.array([l_uv])).T
            j2_luv = np.vander([l_uv], 5)[:, 0:-1] * np.array([5., 4., 3., 2.]) - 1.
            n_uv = idx_uv.sum()
            jacobian[idx_uv] = j2_luv * (rwl[idx_uv]-l_uv).reshape(n_uv, -1) + j1_luv * np.ones(n_uv).reshape(n_uv, -1)

        if has_ir_data:
            l_ir = self.max_reduced_lambda
            j1_lir = (np.vander([l_ir], 6)[:, 0:-2].T - np.array([l_ir])).T
            j2_lir = np.vander([l_ir], 5)[:, 0:-1] * np.array([5., 4., 3., 2.]) - 1.
            n_ir = idx_ir.sum()
            jacobian[idx_ir] = j2_lir * (rwl[idx_ir]-l_ir).reshape(n_ir, -1) + j1_lir * np.ones(n_ir).reshape(n_ir, -1)

        # check for nan's and inf's
        if np.any(np.isnan(r)) or np.any(np.isinf(r)):
            raise ValueError('inf/nan values in color law')

        #
        # if return_jacobian_as_coo_matrix:
        #    jacobian = scipy.sparse.coo_matrix(jacobian)

        return r, jacobian



def check_grad(wl, color_law, p):
    """
    """
    v, jacobian = color_law(wl, p, jac=True)
    dx = 1.E-7
    pp = p.copy()
    df = []
    for i in range(len(pp)):
        pp[i] += dx
        vp, _ = color_law(wl, pp, jac=False)
        df.append((vp-v)/dx)
        pp[i] -= dx
    return jacobian, np.vstack(df).T



#!/usr/bin/env python3

import numpy as np
import scipy.sparse
from bbf.bspline import BSpline
from nacl.sparseutils import kron_product_by_line


if __name__ == '__main__':

    grid = np.linspace(-10., 10., 4)
    basis = BSpline(grid)
    x = np.random.uniform(-10., 10., 4)
    J = basis.eval(x)
    F = np.random.rand(*J.shape)

    K = kron_product_by_line(J, F)



import numpy as np

from nacl.models import constraints

def test_salt2like_constraints(nsn=5, seed=0):
    """
    """
    pass


def test_salt2like_constraint_derivatives(nsn=5, seed=0):
    """
    """
    pass

def test_salt2like_constraints_deactivate(nsn=5, seed=0):
    """
    """
    pass



###########    READ   ###########

##   Use snf_fit to fit the data
##   Use snf_plot to shpw the relevant plots after the fit
##   The .lib files need to be uncommented in the model if they are and also present( not .delete)
##   The hd5 file needs to be in the same directory as this file





import numpy as np
import pylab as pl
import logging
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                    level=logging.INFO)
from nacl import dataset
from nacl.dataset import TrainingDataset
#from nacl.models.salt import SALT2Like, SALT2Eval
#from nacl.models.constraints import SALT2LikeConstraints
#from nacl.models.regularizations import NaClSplineRegularization
from nacl import minimize
from nacl.fit import fit
#import helpers



from nacl.fit import TrainSALT2Like

#from nacl.models.salt2.salt import SALT2Like
#from nacl.models.salt2.constraints import SALT2LikeConstraints
#from nacl.models.salt2.regularizations import NaClSplineRegularization

from nacl.models import salt2
import pandas as pd
from nacl.loglikelihood import LogLikelihood
from nacl.minimize import Minimizer

snf_tds = None

def load():
    global snf_tds
    if snf_tds is None:
        snf_tds = dataset.read_hdf('snf_tds_spline.hd5')
    return snf_tds

def filter_tds(tds, min_wavelength=3600.):
    idx = tds.spectrophotometric_data.wavelength <= min_wavelength
    logging.info(f'invalidating {idx.sum()} measurements outside wl range')
    tds.spectrophotometric_data.valid[idx] = 0
    tds.kill_sne(['PTF12iiq', 'LSQ13aiz', 'PTF11qzq', 'SN2006X'])
    # tds.kill_sne(['PTF12ikt', 'PTF10zdk', 'PTF11bgv'])
    # tds.kill_sne(['SN2004gc', 'PTF11kly', 'SN2005cg'])
    tds.kill_photometric_spectra([1121, 3116, 3307, 1803, 690,
                                   788, 1670, 1056, 1057, 732, 1614,
                                   817, 1224, 1225, 1226, 1227, 1228,
                                  1229, 1230, 1231, 1232, 1233, 1234,
                                  1235, 1236, 1237, 1238, 1239, 1240,
                                  1241, 1242, 1243, 1244, 1245, 1246,
                                  1580, 1854])

    flux = tds.spectrophotometric_data.flux
    fluxerr = tds.spectrophotometric_data.fluxerr
    ey = np.sqrt(fluxerr**2 + 0.05**2 * flux**2)
    tds.spectrophotometric_data.nt['fluxerr'] = ey

    tds.compress()

    return tds

#SNF Fitter

def snf_fit(mu_reg=1., block_names = ['X0', 'X1', 'col', 'tmax', 'M0', 'M1', 'CL'], error=True):
    """
    Fitting the SNFactory data with SALT2 model with the addition of a simple error snake
    (optional) : var = (gamma * v) ** 2
    
    Parameters:
    ----------
    mu_reg : float 
            regularization intensity, default = 1.
    block_names : list
            Parameters to be fitted in the model, default = X0, X1, col, tmax, M0, M1, CL
    error = bool
            Whether or not to include the error snake, default = True
    
    Returns:
    --------
    res : dict
            Dictionnary containing all of the information regarding the minimisation process
            such as the final parameters, the loglikelihood (in minz), etc...
    """
    tds = load()
    tds = filter_tds(tds)
    tm = TrainSALT2Like(tds, mu_reg=mu_reg)
    res = tm.fit(block_names=block_names, fit_error_model=error)
    return res


def snf_mini_fit(tds, mu_reg=1., block_names = ['X0', 'X1', 'col', 'tmax', 'M0', 'M1', 'CL']):
    """
    Testing elements of the fit, do not use for an actual fit for snf
    """
    model = salt2.get_model(tds)
    model.init_from_training_dataset()
    model.init_from_salt2(None, stick_to_original_model=False)
    reg = salt2.get_regularization_prior(model, mu=mu_reg, order=1, check=True)
    cons = salt2.get_constraint_prior(model, linear=True,
                                      mu=1.E6, Mb=-19.5,
                                      check=True)
    # dp = data
    d = {'y':tds.get_all_fluxes(), 'yerr':tds.get_all_fluxerr(), 'bads':tds.get_valid()==0}
    dp = pd.DataFrame(d)
    ll = LogLikelihood(model, reg=[reg], cons=[cons], data=dp)
    ll.pars.release()
    minz = Minimizer(ll)
    r = minz.minimize_lm(p_init=ll.pars.free, lamb=1.E-6, dchi2=10., max_iter=10)
    minz.plot()

    
def train(tds):
    model = SALT2Like(tds,
                      init_from_salt2_file='salt2.npz',
                      init_from_training_dataset=True)
    # we start with a simple light curve fit
    lcfit = fit(model, block_names=['X0', 'X1', 'col', 'tmax'],
                n_iter=10)

    # then, we fit the model itself
    model = SALT2Like(tds, phase_range=(-20, 50.),
                      wl_range=(3000., 10000.),
                      basis_knots=(600, 60),
                      init_from_salt2_file='salt2.npz')
    for block_name in ['X0', 'X1', 'col', 'tmax']:
        model.pars[block_name].full[:] = lcfit['pars'][block_name].full[:]

    # fit the model
    modfit = fit(model, block_names=['M0', 'M1', 'CL'],
                 init_pars=lcfit['pars'])
    p = modfit['pars']

    # then, we release everything
    allfit = fit(model, block_names=['M0', 'M1', 'CL',
                                     'X0', 'X1', 'col', 'tmax'],
                 init_pars=p,
                 n_iter=10)

    return allfit

def plot_results(res):
    model = res['model']
    tds = model.training_dataset
    v = model(model.pars.free, jac=False)


    # residuals
    pl.figure()
    x = np.arange(len(v))
    idx = tds.spectrophotometric_data.valid > 0
    res = tds.get_all_fluxes() - v
    yerr=tds.get_all_fluxerr()
    print(x, idx, res)
    pl.plot(x, res, ls='', color='r', marker='.')
    pl.errorbar(x[idx], res[idx], yerr=tds.get_all_fluxerr()[idx],
                ls='', color='k', marker='.')

    # SN_chi2
    pl.figure()
    wres = res / yerr
    sn_chi2 = np.bincount(tds.spectrophotometric_data.sn_index[idx],
                          wres[idx]**2)
    sn_nmeas = np.bincount(tds.spectrophotometric_data.sn_index[idx])
    rchi2_sn = sn_chi2 / sn_nmeas
    pl.plot(rchi2_sn, 'ro')
    pl.title('SN $\chi^2$')
    pl.xlabel('sn_index$')
    pl.ylabel('SN $\chi^2$')
    
    pl.figure()
    pl.hist(rchi2_sn, bins=40)
    pl.title('SN $\chi^2$/n_meas')
    pl.ylabel('Counts')
    pl.xlabel('SN $\chi^2$/n_meas')
    print('SNe with chi2 > 20')
    isn = np.where(rchi2_sn > 20)[0]
    print(isn)
    print(tds.sn_data.sn_set[isn])
    print('Chi2 = ')
    print(rchi2_sn[isn])

    # spectrum chi2
    pl.figure()
    spec_chi2 = np.bincount(tds.spectrophotometric_data.spec_index[idx],
                            wres[idx]**2)
    spec_nmeas = np.bincount(tds.spectrophotometric_data.spec_index[idx])
    rchi2_spec = spec_chi2 / spec_nmeas
    pl.plot(rchi2_spec, 'ro')
    pl.title('spectrum $\chi^2$')
    pl.xlabel('spec_index')
    pl.ylabel('$\chi^2$')
    
    pl.figure()
    pl.hist(rchi2_spec, bins=80)
    pl.title('spec $\chi^2$/n_meas')
    pl.ylabel('Counts')
    pl.xlabel('spec $\chi^2$/n_meas')
    
    print('spectra with chi2 > 20')
    ispec = np.where(rchi2_spec > 20)[0]
    print(ispec)
    print(tds.spectrophotometric_data.spec_set[ispec])
    #print(tds.spectrophotometric_data.spec_set[ispec])

def plot_spectrum(tds, v, spec):
    pl.figure()
    idx = tds.spectrophotometric_data.spec == spec
    pl.plot(tds.spectrophotometric_data.wavelength[idx],
            tds.spectrophotometric_data.flux[idx], 'r+')
    pl.errorbar(tds.spectrophotometric_data.wavelength[idx],
                tds.spectrophotometric_data.flux[idx],
                yerr=tds.spectrophotometric_data.fluxerr[idx],
                ls='', marker='.', label='data')
    pl.plot(tds.spectrophotometric_data.wavelength[idx], v[idx],
            color='r', lw=2, ls='-', zorder=10,
            label='trained model')
    pl.title(f'spectrum #{spec}', fontsize=16)
    pl.xlabel('observer frame $\lambda\ [\AA]$', fontsize=16)
    pl.ylabel('flux', fontsize=16)
    pl.legend(loc='best')
    
def plot_spectrum_snake(tds, v, spec, var):
    """
    Plotting fitted spectra
    """
    pl.figure()
    idx = tds.spectrophotometric_data.spec == spec
    pl.plot(tds.spectrophotometric_data.wavelength[idx],
            tds.spectrophotometric_data.flux[idx], 'r+')
    pl.errorbar(tds.spectrophotometric_data.wavelength[idx],
                tds.spectrophotometric_data.flux[idx],
                yerr=tds.spectrophotometric_data.fluxerr[idx],
                ls='', marker='.', label='data')
    pl.plot(tds.spectrophotometric_data.wavelength[idx], v[idx],
            color='r', lw=2, ls='-', zorder=10,
            label='trained model')
    pl.fill_between(tds.spectrophotometric_data.wavelength[idx], 
                    v[idx]-np.sqrt(var[idx]), v[idx]+np.sqrt(var[idx]), alpha = 0.5)
    pl.title(f'spectrum #{spec}', fontsize=16)
    pl.xlabel('observer frame $\lambda\ [\AA]$', fontsize=16)
    pl.ylabel('flux', fontsize=16)
    pl.legend(loc='best')

def plot_stacked_residuals(tds, model, pars, v=None):
    """
    """
    if v is None:
        v = model(pars, jac=False)

    wl = tds.spectrophotometric_data.wavelength / (1. + tds.spectrophotometric_data.z)
    ph = (tds.spectrophotometric_data.mjd - pars['tmax'].full[tds.spectrophotometric_data.sn_index]) / (1 + tds.spectrophotometric_data.z)

    pl.figure(figsize=(14, 10))
    idx = tds.get_valid() > 0
    res = tds.get_all_fluxes() - v
    wres = res / tds.get_all_fluxerr()
    pl.hexbin(wl[idx], ph[idx], wres[idx], vmin=-2.5, vmax=2.5)
    pl.title('weighted residuals', fontsize=16)
    pl.xlabel('restframe $\lambda [\AA]$', fontsize=16)
    pl.ylabel('restframe phase [days]', fontsize=16)
    pl.colorbar()
    
def plot_stacked_residuals_snake(tds, model, pars, var, v=None):
    """
    Pulls with snake error
    """
    if v is None:
        v = model(pars, jac=False)

    wl = tds.spectrophotometric_data.wavelength / (1. + tds.spectrophotometric_data.z)
    ph = (tds.spectrophotometric_data.mjd - pars['tmax'].full[tds.spectrophotometric_data.sn_index]) / (1 + tds.spectrophotometric_data.z)

    pl.figure(figsize=(14, 10))
    idx = tds.get_valid() > 0
    res = tds.get_all_fluxes() - v
    wres = res / (tds.get_all_fluxerr() + np.sqrt(var))
    pl.hexbin(wl[idx], ph[idx], wres[idx], vmin=-2.5, vmax=2.5)
    pl.title('pulls with simple snake', fontsize=16)
    pl.xlabel('restframe $\lambda [\AA]$', fontsize=16)
    pl.ylabel('restframe phase [days]', fontsize=16)
    pl.colorbar()

def plot_chi2(tds, model, v=None):
    if v is None:
        v = model(model.pars, jac=False)
    wl = tds.spectrophotometric_data.wavelength / (1. + tds.spectrophotometric_data.z)
    ph = (tds.spectrophotometric_data.mjd - model.pars['tmax'].full[tds.spectrophotometric_data.sn_index]) / (1 + tds.spectrophotometric_data.z)

    pl.figure(figsize=(14, 10))
    idx = tds.get_valid() > 0
    res = tds.get_all_fluxes() - v
    wres = res / tds.get_all_fluxerr()
    chi2, _, _ = np.histogram2d(wl, ph, bins=(400,50), weights=wres**2)
    cc, _, _ = np.histogram2d(wl, ph, bins=(400,50))
    rchi2 = chi2 / cc
    pl.imshow(rchi2.T[::-1,:], vmin=0, vmax=10,
              aspect='auto',
              interpolation='none',
              extent=[wl.min(), wl.max(), ph.min(), ph.max()])
    pl.title('local $\chi^2$', fontsize=16)
    pl.xlabel('restframe $\lambda [\AA]$', fontsize=16)
    pl.ylabel('restframe phase [days]', fontsize=16)
    pl.colorbar()

def plot_chi2_snake(tds, model, var, v=None):
    """
    Chi2 with snake error
    """
    if v is None:
        v = model(model.pars, jac=False)
    wl = tds.spectrophotometric_data.wavelength / (1. + tds.spectrophotometric_data.z)
    ph = (tds.spectrophotometric_data.mjd - model.pars['tmax'].full[tds.spectrophotometric_data.sn_index]) / (1 + tds.spectrophotometric_data.z)

    pl.figure(figsize=(14, 10))
    idx = tds.get_valid() > 0
    res = tds.get_all_fluxes() - v
    wres = res / (tds.get_all_fluxerr() + np.sqrt(var))
    chi2, _, _ = np.histogram2d(wl, ph, bins=(400,50), weights=wres**2)
    cc, _, _ = np.histogram2d(wl, ph, bins=(400,50))
    rchi2 = chi2 / cc
    pl.imshow(rchi2.T[::-1,:], vmin=0, vmax=10,
              aspect='auto',
              interpolation='none',
              extent=[wl.min(), wl.max(), ph.min(), ph.max()])
    pl.title('local $\chi^2$', fontsize=16)
    pl.xlabel('restframe $\lambda [\AA]$', fontsize=16)
    pl.ylabel('restframe phase [days]', fontsize=16)
    pl.colorbar()

def snf_plotter(res):
    """
    Plots the relevant plots after a fit and some spectra.
    
    Parameters:
    ----------
    res : dict
    
    """
    minz = res['minz']
    ll = minz.log_likelihood
    model = ll.model
    variance_model = ll.variance_model
    tds = model.training_dataset
    
    v = model(model.pars, jac=False)
    var = variance_model(ll.pars, jac=False)
    spec0 = 0
    spec1 = 2642
    spec2 = 518
    
    plot_stacked_residuals_snake(tds, model, var, v)
    plot_chi2_snake(tds, model, var, v=None)
    plot_spectrum_snake(tds, v, spec0, var)
    plot_spectrum_snake(tds, v, spec1, var)
    plot_spectrum_snake(tds, v, spec2, var)
    

def train_orig(tds):
    """Fit a NaCl model on the TrainingDataset
    """
    # we start with a light curve fit
    model = SALT2Like(tds, init_from_salt2_file='salt2.npz')
    model.init_from_training_dataset()

    model.pars.fix()
    model.pars['X0'].release()
    model.pars['X1'].release()
    model.pars['col'].release()
    model.pars['tmax'].release()

    # reg = NaClSplineRegularization(model, to_regularize=['M0'],
    #                                order=0, mu=1.)
    wres = minimize.WeightedResiduals(model)
    chi2 = minimize.LogLikelihood(wres)
    minz = minimize.Minimizer(chi2)
    ret = minz.minimize(model.pars.free, dchi2_stop=0.1)

    # now, we fix everything but M0 (and M1)
    for block_name in ['X0', 'X1', 'col', 'tmax']:
        model.pars[block_name].full[:] = ret['pars'][block_name].full[:]
    model.pars.fix()
    for block_name in ['M0', 'M1', 'CL']:
        model.pars[block_name].release()

    reg = NaClSplineRegularization(model, to_regularize=['M0', 'M1'],
                                    order=0, mu=1.)
    wres = minimize.WeightedResiduals(model)
    chi2 = minimize.LogLikelihood(wres, reg=[reg])
    minz = minimize.Minimizer(chi2)
    ret = minz.minimize(model.pars.free)

    return ret, minz

def compare_pars(tds, model):
    """
    compares the parameters found in the IDR and after the NaCl fit
    
    model : SALT2Like result after fit(tds)
    """
    
    x0_tds = tds.sn_data.x0
    x1_tds = tds.sn_data.x1
    c_tds = tds.sn_data.col
    tmax_tds = tds.sn_data.tmax
    z = tds.sn_data.z
    
    x0_mod = model.pars['X0'].full[tds.sn_data.sn_index]
    x1_mod = model.pars['X1'].full[tds.sn_data.sn_index]
    c_mod = model.pars['col'].full[tds.sn_data.sn_index]
    tmax_mod = model.pars['tmax'].full[tds.sn_data.sn_index]
    
    pl.figure()
    pl.scatter(x0_tds, x0_mod)
    pl.title('X0 NaCl vs X0 IDR')
    pl.xlabel('X0 IDR')
    pl.ylabel('X0 NaCl')
    pl.show()
    
    pl.figure()
    pl.scatter(x1_tds, x1_mod)
    pl.title('X1 NaCl vs X1 IDR')
    pl.xlabel('X1 IDR')
    pl.ylabel('X1 NaCl')
    pl.show()
    
    pl.figure()
    pl.scatter(c_tds, c_mod)
    pl.title('col NaCl vs col IDR')
    pl.xlabel('col IDR')
    pl.ylabel('col NaCl')
    pl.show()
    
    pl.figure()
    pl.scatter(tmax_tds, tmax_mod)
    pl.title('tmax NaCl vs tmax IDR')
    pl.xlabel('tmax IDR')
    pl.ylabel('tmax NaCl')
    pl.show()
    
    pl.figure()
    pl.scatter(z, x0_mod/x0_tds, color='red')
    pl.title('X0 NaCl/X0 IDR vs z')
    pl.xlabel('z')
    pl.ylabel('X0 NaCl/X0 IDR')
    pl.show()

    pl.figure()
    pl.scatter(z, x1_mod/x1_tds, color='red')
    pl.title('X1 NaCl/X1 IDR vs z')
    pl.xlabel('z')
    pl.ylabel('X1 NaCl/X1 IDR')
    pl.show()
    
    pl.figure()
    pl.scatter(z, c_mod/c_tds, color='red')
    pl.title('col NaCl/col IDR vs z')
    pl.xlabel('z')
    pl.ylabel('col NaCl/col IDR')
    pl.show()
    
    pl.figure()
    pl.scatter(z, tmax_mod/tmax_tds, color='red')
    pl.title('tmax NaCl/tmax IDR vs z')
    pl.xlabel('z')
    pl.ylabel('tmax NaCl/tmax IDR')
    pl.show()

    pl.figure()
    pl.scatter(z, x1_mod-x1_tds, color='red')
    pl.title('X1 NaCl-X1 IDR vs z')
    pl.xlabel('z')
    pl.ylabel('X1 NaCl-X1 IDR')
    pl.show()
    
    pl.figure()
    pl.scatter(z, c_mod-c_tds, color='red')
    pl.title('col NaCl-col IDR vs z')
    pl.xlabel('z')
    pl.ylabel('col NaCl-col IDR')
    pl.show()
    
        
    pl.figure()
    pl.scatter(z, np.sqrt( 1/x0_mod ), label='NaCl' )
    pl.scatter(z, np.sqrt( 1/x0_tds ), label='IDR' )
    pl.xlabel('z')
    pl.ylabel(r'$\sqrt{1/X0}$')
    pl.legend()
    pl.show()
    
    pl.figure()
    pl.scatter(z, 1/x0_mod, label='NaCl')
    pl.scatter(z, 1/x0_tds, label='IDR')
    pl.xlabel('z')
    pl.ylabel('1/X0')
    pl.yscale('log')
    pl.legend()
    pl.show()
